"""an executable to upload auxiliary files and plots for GraceDB"""

# Copyright (C) 2019      Alexander Pace, Kipp Cannon, Chad Hanna, Drew Keppel
# Copyright (C) 2020      Patrick Godwin, Cody Messick
# Copyright (C) 2024-2025 Yun-Jing Huang

__usage__ = "sgnl-ll-inspiral-event-plotter [--options]"

# -------------------------------------------------
#                   Preamble
# -------------------------------------------------

import copy
import http.client
import io
import json
import logging
import math
import os
import time
from argparse import ArgumentParser
from collections import OrderedDict
from enum import Enum
from xml.sax import SAXParseException

import numpy
from igwn_ligolw import ligolw, lsctables
from igwn_ligolw import utils as ligolw_utils
from igwn_ligolw.array import use_in as array_use_in
from igwn_ligolw.ligolw import Param as ligolw_param
from igwn_ligolw.param import use_in as param_use_in
from lal import LIGOTimeGPS, series
from ligo.gracedb.rest import DEFAULT_SERVICE_URL as DEFAULT_GRACEDB_URL
from ligo.gracedb.rest import GraceDb, HTTPError
from ligo.scald import utils
from strike.stats.likelihood_ratio import LnLikelihoodRatio

from sgnl.plots.util import set_matplotlib_cache_directory

set_matplotlib_cache_directory()  # FIXME: I don't know if this needs to be done here

from strike.plots.stats import (  # noqa: E402
    plot_dtdphi,
    plot_horizon_distance_vs_time,
    plot_rates,
    plot_snr_chi_pdf,
)

from sgnl import events, svd_bank  # noqa: E402 I003
from sgnl.gracedb import FakeGracedbClient, upload_fig  # noqa: E402
from sgnl.plots import psd as plotpsd  # noqa: E402

# from strike.plots.stats import far as plotfar  # noqa: E402

import matplotlib  # noqa: E402 I001 isort:skip

matplotlib.rcParams.update(
    {
        "font.size": 10.0,
        "axes.titlesize": 10.0,
        "axes.labelsize": 10.0,
        "xtick.labelsize": 8.0,
        "ytick.labelsize": 8.0,
        "legend.fontsize": 8.0,
        "figure.dpi": 100,
        "savefig.dpi": 100,
        "text.usetex": False,
    }
)
from matplotlib import figure  # noqa: E402
from matplotlib.backends.backend_agg import (  # noqa: E402
    FigureCanvasAgg as FigureCanvas,
)

matplotlib.use("agg")
from matplotlib import pyplot as plt  # noqa: E402


# -------------------------------------------------
#                Content Handler
# -------------------------------------------------
@lsctables.use_in
@array_use_in
@param_use_in
class ligolwcontenthandler(ligolw.LIGOLWContentHandler):
    pass


# -------------------------------------------------
#                   Functions
# -------------------------------------------------


def parse_command_line():

    parser = ArgumentParser(usage=__usage__, description=__doc__)
    parser.add_argument(
        "-v", "--verbose", default=False, action="store_true", help="Be verbose."
    )
    parser.add_argument(
        "--tag",
        metavar="string",
        default="test",
        help="Sets the name of the tag used. Default = 'test'",
    )
    parser.add_argument(
        "--max-event-time",
        type=int,
        default=7200,
        help="Maximum time to keep around an event. Default = 2 hours.",
    )
    parser.add_argument(
        "--processing-cadence",
        type=float,
        default=0.1,
        help="Rate at which the event plotter acquires and processes data. Default = "
        "0.1 seconds.",
    )
    parser.add_argument(
        "--request-timeout",
        type=float,
        default=0.2,
        help="Timeout for requesting messages from a topic. Default = 0.2 seconds.",
    )
    parser.add_argument(
        "--kafka-server",
        metavar="string",
        help="Sets the server url that the kafka topic is hosted on. Required.",
    )
    parser.add_argument(
        "--upload-topic",
        metavar="string",
        help="Sets the input kafka topic to get uploaded event info from. Required.",
    )
    parser.add_argument(
        "--ranking-stat-topic",
        metavar="string",
        help="Sets the input kafka topic to get ranking stat info from. Required.",
    )
    parser.add_argument(
        "--gracedb-service-url",
        metavar="url",
        default=DEFAULT_GRACEDB_URL,
        help="Override default GracedB service url (optional, default is {}).".format(
            DEFAULT_GRACEDB_URL
        ),
    )
    parser.add_argument(
        "--max-snr",
        metavar="SNR",
        type=float,
        default=200.0,
        help="Set the upper bound of the SNR ranges in plots (default = 200).",
    )
    parser.add_argument(
        "--format",
        default="png",
        help="Set file format by selecting the extention (default = 'png').",
    )
    parser.add_argument(
        "--output-path",
        metavar="PATH",
        help="Write local copies of the plots to this directory (default = don't).",
    )
    parser.add_argument(
        "--no-upload",
        action="store_true",
        help="Disable upload of plots to gracedb, e.g., for testing new plots.",
    )
    parser.add_argument(
        "--plot",
        action="append",
        help=(
            "Plots to make. Options are: "
            "1. RANKING_DATA: upload: ranking_data.xml.gz. "
            "2. RANKING_PLOTS: upload background plots from the ranking statistic data."
            "3. SNR_PLOTS: upload SNR time series plots. "
            "4. PSD_PLOTS: upload PSD plots. "
            "5. DTDPHI_PLOTS: upload DTDPHI plots. "
            " Can be given multiple times. If not specified, all plots are made."
        ),
    )

    options = parser.parse_args()

    if options.no_upload and options.output_path is None:
        raise ValueError("--no-upload without setting --ouput-path disables all output")

    return options


def load_rankingstat_xml_with_retries(filename, logger):
    # The rankingstat is not guaranteed to be available on disk immediately
    # after getting written. Try 3 times, with 1s sleep in between.
    for tries in range(3):
        try:
            xmldoc = ligolw_utils.load_filename(
                filename, contenthandler=LnLikelihoodRatio.LIGOLWContentHandler
            )
            break
        except (SAXParseException, OSError) as e:
            if tries < 2:
                logger.warning(
                    "Failed to load rankingstat located at %s. Trying again.", filename
                )
                time.sleep(1)
            else:
                raise IOError(
                    f"Could not load rankingstat located at {filename} after 3 tries. "
                    f"Failed with error:\n{e}"
                )
    return xmldoc


# -------------------------------------------------
#                    Classes
# -------------------------------------------------


class Plots(Enum):
    RANKING_DATA = 1
    RANKING_PLOTS = 2
    SNR_PLOTS = 3
    PSD_PLOTS = 4
    DTDPHI_PLOTS = 5


class EventPlotter(events.EventProcessor):
    """
    manages plotting and file uploading for incoming events.
    """

    _name = "event_plotter"

    def __init__(
        self,
        kafka_server: str,
        logger,
        output_path: str,
        plot: list[str],
        upload_topic: str,
        ranking_stat_topic: str,
        format: str = "png",
        gracedb_service_url: str = DEFAULT_GRACEDB_URL,
        max_event_time: int = 7200,
        max_snr: float = 200.0,
        no_upload: bool = False,
        processing_cadence: float = 0.1,
        request_timeout: float = 0.2,
        tag: str = "test",
    ):
        self.logger = logger
        self.logger.info("setting up event plotter...")

        self.upload_topic = f"sgnl.{tag}.{upload_topic}"
        self.ranking_stat_topic = f"sgnl.{tag}.{ranking_stat_topic}"

        plot_string = "-".join([str(Plots[plot].value) for plot in plot])

        is_injection_job = upload_topic == "inj_uploads"
        heartbeat_topic = (
            f"sgnl.{tag}.event_plotter_heartbeat"
            if not is_injection_job
            else f"sgnl.{tag}.inj_event_plotter_heartbeat"
        )

        events.EventProcessor.__init__(
            self,
            process_cadence=processing_cadence,
            request_timeout=request_timeout,
            kafka_server=kafka_server,
            input_topic=[self.upload_topic, self.ranking_stat_topic],
            tag=f"{tag}-{plot_string}",
            send_heartbeats=True,
            heartbeat_cadence=60.0,
            heartbeat_topic=heartbeat_topic,
        )

        # initialize timing options
        self.max_event_time = max_event_time
        self.retries = 5
        self.retry_delay = 1

        # initialize gracedb client
        if gracedb_service_url.startswith("file"):
            self.client = FakeGracedbClient(gracedb_service_url)
        else:
            self.client = GraceDb(gracedb_service_url)

        # initialize event storage
        self.events: OrderedDict = OrderedDict()

        # initialize plotting options
        self.to_upload = plot
        self.max_snr = max_snr
        self.format = format
        self.output_path = output_path
        self.no_upload = no_upload

    def ingest(self, message):
        """
        parse a message from a kafka topic
        """
        payload = json.loads(message.value())

        time = LIGOTimeGPS(payload["time"], payload["time_ns"])
        coinc_fileobj = io.BytesIO(payload["coinc"].encode("utf-8"))
        psd_fileobj = copy.copy(coinc_fileobj)
        xmldoc = ligolw_utils.load_fileobj(
            coinc_fileobj, contenthandler=ligolwcontenthandler
        )
        coinc_fileobj.close()
        sngl_inspiral_table = lsctables.SnglInspiralTable.get_table(xmldoc)
        bank_bin = "{:04}".format(int(sngl_inspiral_table[0].Gamma1))
        # No guarantee that the coinc event id will be unique between
        # bins, so use int(time) and the bank bin as an identifier,
        # which should be unique as only one event / second / bin may
        # be uploaded
        event_key = "{}_{}".format(payload["time"], bank_bin)

        if event_key not in self.events:
            self.logger.info("found new event at %s from bin %s", time, bank_bin)
            self.events[event_key] = self.new_event(time, bank_bin)

        # ranking stat
        if message.topic() == self.ranking_stat_topic:
            self.events[event_key]["ranking_data_path"] = payload["ranking_data_path"]
            # we'll just take the xmldoc from the preferred event, which will be
            # identical
            xmldoc.unlink()

        # preferred event
        elif message.topic() == self.upload_topic:
            self.logger.info("GID %s", payload["gid"])
            self.events[event_key]["gid"] = payload["gid"]
            self.events[event_key]["coinc"] = xmldoc
            self.events[event_key]["psd"] = ligolw_utils.load_fileobj(
                psd_fileobj, contenthandler=series.PSDContentHandler
            )
            self.events[event_key]["snr_optimized"] = (
                False if "snr_optimized" not in payload else payload["snr_optimized"]
            )
        psd_fileobj.close()

    def new_event(self, time, bank_bin):
        """
        returns the structure that defines an event
        """
        return {
            "time": time,
            "bin": bank_bin,
            "coinc": None,
            "gid": None,
            "psd": None,
            "ranking_data_path": None,
            "uploaded": {upload: False for upload in self.to_upload},
        }

    def handle(self):
        """
        handle aux data and plot uploading, clearing out
        old events as necessary.
        """
        for event in self.events.values():
            uploaded = event["uploaded"]
            if event["gid"]:
                if (
                    "RANKING_DATA" in uploaded
                    and not uploaded["RANKING_DATA"]
                    and event["ranking_data_path"]
                ):
                    self.upload_ranking_data(event)
                    uploaded["RANKING_DATA"] = True
                if (
                    "RANKING_PLOTS" in uploaded
                    and not uploaded["RANKING_PLOTS"]
                    and event["ranking_data_path"]
                ):
                    self.upload_ranking_plots(event)
                    uploaded["RANKING_PLOTS"] = True
                if (
                    "PSD_PLOTS" in uploaded
                    and not uploaded["PSD_PLOTS"]
                    and event["psd"]
                ):
                    self.upload_psd_plots(event)
                    uploaded["PSD_PLOTS"] = True
                if "SNR_PLOTS" in uploaded and not uploaded["SNR_PLOTS"]:
                    self.upload_snr_plots(event)
                    uploaded["SNR_PLOTS"] = True
                if (
                    "DTDPHI_PLOTS" in uploaded
                    and not uploaded["DTDPHI_PLOTS"]
                    and event["ranking_data_path"]
                ):
                    self.upload_dtdphi_plots(event)
                    uploaded["DTDPHI_PLOTS"] = True

        # clean out events once all plots are uploaded
        # and clean out old events
        current_time = utils.gps_now()
        for event_key in list(self.events.keys()):
            event = self.events[event_key]
            if (
                all(event["uploaded"].values())
                or current_time - event["time"] >= self.max_event_time
            ):
                self.logger.info(
                    "removing event from %s and bin %s", event["time"], event["bin"]
                )
                if event["coinc"] is not None:
                    self.logger.info(
                        "Did not receive path of ranking data file associated with "
                        "event from %s and bin %s",
                        event["time"],
                        event["bin"],
                    )
                    event["coinc"].unlink()
                    event["psd"].unlink()
                self.events.pop(event_key)

    def upload_file(self, message, filename, tag, contents, graceid):
        """
        upload a file to gracedb
        """
        self.logger.info("posting %s to gracedb ID %s", filename, graceid)
        for attempt in range(1, self.retries + 1):
            try:
                resp = self.client.writeLog(
                    graceid,
                    message,
                    filename=filename,
                    filecontents=contents,
                    tagname=tag,
                )
            except HTTPError:
                self.logger.exception("HTTPError")
            else:
                if resp.status == http.client.CREATED:
                    break
            self.logger.info(
                "gracedb upload of %s for ID %s " "failed on attempt %d/%d",
                filename,
                graceid,
                attempt,
                self.retries,
            )
            time.sleep(numpy.random.lognormal(math.log(self.retry_delay), 0.5))
        else:
            self.logger.warning(
                "gracedb upload of %s for ID %s failed", filename, graceid
            )
            return False

    def upload_ranking_data(self, event):
        ranking_fobj = io.BytesIO()
        gz = "gz" if event["ranking_data_path"].endswith("gz") else False
        ligolw_utils.write_fileobj(
            load_rankingstat_xml_with_retries(event["ranking_data_path"], self.logger),
            ranking_fobj,
            compress=gz,
        )
        self.upload_file(
            "ranking statistic PDFs",
            "ranking_data.xml.gz",
            "ranking_statistic",
            ranking_fobj.getvalue(),
            event["gid"],
        )
        ranking_fobj.close()

    def upload_ranking_plots(self, event):
        self.logger.info("Begin plotting ranking plots for %s...", event["gid"])
        # load all of the information needed to generate plots
        sngl_inspirals = dict(
            (row.ifo, row)
            for row in lsctables.SnglInspiralTable.get_table(event["coinc"])
        )
        coinc_event_table = lsctables.CoincTable.get_table(event["coinc"])
        try:
            (coinc_event,) = coinc_event_table
        except ValueError:
            raise ValueError("document does not contain exactly one candidate")

        xmldoc = load_rankingstat_xml_with_retries(
            event["ranking_data_path"], self.logger
        )
        rankingstat = LnLikelihoodRatio.from_xml(xmldoc)
        rankingstat.finish()
        # FIXME: this is needed for the likelihood_ratio_ccdf plot, do we want this?
        # rankingstatpdf = far.RankingStatPDF.from_xml(xmldoc, "strike_rankingstatpdf")
        # fapfar = far.FAPFAR(rankingstatpdf.new_with_extinction())

        # generate and upload plots
        for plot_type in ["background_pdf", "injection_pdf", "zero_lag_pdf", "LR"]:
            for instrument in rankingstat.instruments:
                # for chi_type in ("chi", "bankchi"):
                if instrument in sngl_inspirals:
                    # place marker on plot
                    if sngl_inspirals[instrument].snr >= 4.0:
                        snr = sngl_inspirals[instrument].snr
                        # chisq = getattr(
                        #     sngl_inspirals[instrument],
                        #     "%ssq" % chi_type.replace("bank", "bank_"),
                        # )
                        chisq = sngl_inspirals[instrument].chisq
                    else:
                        snr = None
                        chisq = None
                    fig = plot_snr_chi_pdf(
                        rankingstat.terms["P_of_SNR_chisq"],
                        instrument,
                        plot_type,
                        self.max_snr,
                        event_snr=snr,
                        event_chisq=chisq,
                    )
                else:
                    # no sngl for this instrument
                    fig = plot_snr_chi_pdf(
                        rankingstat.terms["P_of_SNR_chisq"],
                        instrument,
                        plot_type,
                        self.max_snr,
                    )
                if fig is not None:
                    filename = "{}_{}_{}_snrchi.{}".format(
                        event["gid"], instrument, plot_type, self.format
                    )
                    if not self.no_upload:
                        upload_fig(
                            fig,
                            self.client,
                            event["gid"],
                            filename=filename,
                            log_message="%s SNR, chisq PDF" % (instrument),
                            tagname="background",
                        )
                    if self.output_path is not None:
                        filename = os.path.join(self.output_path, filename)
                        self.logger.info("writing %s ...", filename)
                        fig.savefig(filename)
                    plt.close(fig)

        # fig = plot_likelihood_ratio_ccdf(
        #     fapfar,
        #     (
        #       0.0,
        #       max(40.0, coinc_event.likelihood - coinc_event.likelihood % 5.0 + 5.0),
        #     ),
        #     ln_likelihood_ratio_markers=(coinc_event.likelihood,),
        # )
        # filename = "{}_likehoodratio_ccdf.{}".format(event["gid"], self.format)
        # if not self.no_upload:
        #     upload_fig(
        #         fig,
        #         self.client,
        #         event["gid"],
        #         filename=filename,
        #         log_message="Likelihood Ratio CCDF",
        #         tagname="background",
        #     )
        # if self.output_path is not None:
        #     filename = os.path.join(self.output_path, filename)
        #     self.logger.info("writing %s ...", filename)
        #     fig.savefig(filename)
        # plt.close(fig)

        fig = plot_horizon_distance_vs_time(
            rankingstat, (event["time"] - 14400.0, event["time"]), tref=event["time"]
        )
        filename = "{}_horizon_distances.{}".format(event["gid"], self.format)
        if not self.no_upload:
            upload_fig(
                fig,
                self.client,
                event["gid"],
                filename=filename,
                log_message="Horizon Distances",
                tagname="psd",
            )
        if self.output_path is not None:
            filename = os.path.join(self.output_path, filename)
            self.logger.info("writing %s ...", filename)
            fig.savefig(filename)
        plt.close(fig)

        fig = plot_rates(rankingstat)
        filename = "{}_rates.{}".format(event["gid"], self.format)
        if not self.no_upload:
            upload_fig(
                fig,
                self.client,
                event["gid"],
                filename=filename,
                log_message="Instrument combo rates",
                tagname="background",
            )
        if self.output_path is not None:
            filename = os.path.join(self.output_path, filename)
            self.logger.info("writing %s ...", filename)
            fig.savefig(filename)
        plt.close(fig)
        self.logger.info("finished processing ranking data plots for %s", event["gid"])

    def upload_psd_plots(self, event):
        psds = series.read_psd_xmldoc(event["psd"])
        if psds is None:
            self.logger.info("Could not get_psds, exiting loop")
            return

        #
        # PSD plot
        #

        fig = plotpsd.plot_psds(psds, event["coinc"], plot_width=800)
        fig.tight_layout()

        filename = "{}_psd.{}".format(event["gid"], self.format)
        if self.no_upload:
            self.logger.info("writing %s ...", filename)
            fig.savefig(filename)
        else:
            upload_fig(
                fig,
                self.client,
                event["gid"],
                filename=filename,
                log_message="strain spectral density plot",
                tagname="psd",
            )
        plt.close(fig)

        #
        # Cumulative SNRs plot
        #

        fig = plotpsd.plot_cumulative_snrs(psds, event["coinc"], plot_width=800)
        fig.tight_layout()

        filename = "{}_cumulative_snrs.{}".format(event["gid"], self.format)
        if self.no_upload:
            self.logger.info("writing %s ...", filename)
            fig.savefig(filename)
        else:
            upload_fig(
                fig,
                self.client,
                event["gid"],
                filename=filename,
                log_message="cumulative SNRs plot",
                tagname="psd",
            )
        plt.close(fig)

        self.logger.info("finished processing psd plot for %s", event["gid"])

    def upload_snr_plots(self, event):
        # create two dicts keyed by event id: the first dict contains
        # COMPLEX8TimeSeries which contain the snr time series, the second dict
        # contains the template row
        timeseries_ligolw_dict = dict(
            (
                ligolw_param.get_param(elem, "event_id").value,
                series.parse_COMPLEX8TimeSeries(elem),
            )
            for elem in event["coinc"].getElementsByTagName(ligolw.LIGO_LW.tagName)
            if elem.hasAttribute("Name") and elem.Name == "COMPLEX8TimeSeries"
        )
        eventid_trigger_dict = dict(
            (row.event_id, row)
            for row in lsctables.SnglInspiralTable.get_table(event["coinc"])
        )

        # we don't have an autocorrelation series for an snr optimized event, so don't
        # plot it
        plot_autocorrelation = (
            False if "snr_optimized" in event and event["snr_optimized"] else True
        )

        # Parse the bank files
        # NOTE This assumes --svd-bank will also be provided once in the
        # ProcessParamsTable
        if plot_autocorrelation:
            self.logger.info("Reading svd bank...")
            bank_files = [
                row.value
                for row in lsctables.ProcessParamsTable.get_table(event["coinc"])
                if row.param == "--svd-bank" and "%04d" % int(event["bin"]) in row.value
            ]
            self.logger.info("Got process params table...")
            svd_bank_string = ",".join(
                [
                    f'{os.path.basename(file).split("-")[0]}:{file}'
                    for file in bank_files
                ]
            )

            banks = svd_bank.parse_bank_files(
                svd_bank.parse_svdbank_string(svd_bank_string), verbose=False
            )
            self.logger.info("Parsed bank files...")

            #
            # Find the template (to retrieve the autocorrelation later)
            #
            banknum = None
            for i, bank in enumerate(list(banks.values())[0]):
                for j, row in enumerate(bank.sngl_inspiral_table):
                    # The templates should all have the same template_id, so just grab
                    # one
                    if row.Gamma0 == list(eventid_trigger_dict.values())[0].Gamma0:
                        banknum = i
                        tmpltnum = j
                        break
                if banknum is not None:
                    break

            if banknum is None:
                raise ValueError(
                    "The svd banks in the process params table do not contain the "
                    "template the event was found with"
                )
            self.logger.info("Finish reading svd bank...")

        #
        # Plot the time series and the expected snr
        #
        fig = figure.Figure()
        FigureCanvas(fig)

        self.logger.info("Begin plotting snr time series...")
        zero_pad = 4
        for i, (eventid, complex8timeseries) in enumerate(
            timeseries_ligolw_dict.items()
        ):
            ifo = eventid_trigger_dict[eventid].ifo
            autocorr_length = complex8timeseries.data.length

            # add zero pad as safety in case the peak offset is not the center of snr
            # timeseries
            time = numpy.linspace(
                float(complex8timeseries.epoch) - zero_pad * complex8timeseries.deltaT,
                float(complex8timeseries.epoch)
                + (autocorr_length + zero_pad - 1) * complex8timeseries.deltaT,
                autocorr_length + zero_pad * 2,
            )
            complex_snr_timeseries = numpy.concatenate(
                [
                    numpy.zeros(zero_pad),
                    complex8timeseries.data.data,
                    numpy.zeros(zero_pad),
                ]
            )
            if plot_autocorrelation:
                auto = numpy.concatenate(
                    [
                        numpy.zeros(zero_pad),
                        banks[ifo][banknum].autocorrelation_bank[tmpltnum],
                        numpy.zeros(zero_pad),
                    ]
                )

            peakoffset = numpy.argmin(abs(time - eventid_trigger_dict[eventid].end))
            phase = numpy.angle(complex_snr_timeseries[peakoffset])
            snr = (complex_snr_timeseries * numpy.exp(-1.0j * phase)).real
            snrsigma = numpy.sqrt(2)
            peaktime = time[peakoffset]
            time -= peaktime
            maxsnr = snr.max()

            lo_idx = int(peakoffset - (autocorr_length - 1) / 2)
            hi_idx = int(peakoffset + (autocorr_length + 1) / 2 + 1)

            ax = fig.add_subplot(len(timeseries_ligolw_dict.items()), 1, i + 1)
            ax.fill_between(
                time[lo_idx:hi_idx],
                snr[lo_idx:hi_idx] - snrsigma,
                snr[lo_idx:hi_idx] + snrsigma,
                color="0.75",
            )
            ax.plot(
                time[lo_idx:hi_idx],
                snr[lo_idx:hi_idx],
                "k",
                label=r"$\mathrm{Measured}\,\rho(t)$",
            )
            if plot_autocorrelation:
                ax.plot(
                    time[lo_idx:hi_idx],
                    auto.real[lo_idx:hi_idx] * maxsnr,
                    "b--",
                    label=r"$\mathrm{Scaled\,Autocorrelation}$",
                )
            ax.set_xlim(time[lo_idx], time[hi_idx])
            ax.set_ylabel(r"$\mathrm{{{}}}\,\rho(t)$".format(ifo))
            ax.set_xlabel(r"$\mathrm{{Time\,from\,{}}}$".format(peaktime))
            ax.legend(loc="upper right")
            ax.grid()

        fig.tight_layout()
        filename = "{}_snrtimeseries.{}".format(event["gid"], self.format)

        if not self.no_upload:
            self.logger.info("writing %s ...", filename)
            upload_fig(
                fig,
                self.client,
                event["gid"],
                filename=filename,
                log_message="SNR time series",
                tagname="background",
            )

        if self.output_path is not None:
            filename = os.path.join(self.output_path, filename)
            self.logger.info("writing %s ...", filename)
            fig.savefig(filename)
        plt.close(fig)

        self.logger.info(
            "finished processing SNR time series plot for %s", event["gid"]
        )

    def upload_dtdphi_plots(self, event):
        self.logger.info("Begin plotting dtdphi plots for %s...", event["gid"])
        sngl_inspiral_table = lsctables.SnglInspiralTable.get_table(event["coinc"])
        offset_vectors = lsctables.TimeSlideTable.get_table(event["coinc"]).as_dict()
        assert (
            len(offset_vectors) == 1
        ), "the time slide table has to have exactly one time-slide entry."

        offset_vector = offset_vectors[list(offset_vectors)[0]]

        rankingstat = LnLikelihoodRatio.from_xml(
            load_rankingstat_xml_with_retries(event["ranking_data_path"], self.logger)
        )
        rankingstat.finish()  # FIXME: is this needed?

        # Map sgnl_inspiral rows to trigger columns
        # FIXME: should this be in strike?
        triggers = []
        for row in sngl_inspiral_table:
            if row.chisq is not None:  # FIXME: sometimes chisq is None breaks the code
                triggers.append(
                    {
                        "__trigger_id": row.event_id,
                        "_filter_id": row.template_id,
                        "ifo": row.ifo,
                        "time": row.end,
                        "snr": row.snr,
                        "chisq": row.chisq,
                        "phase": row.coa_phase,
                        "Gamma2": row.Gamma2,
                        "template_duration": row.template_duration,
                    }
                )
        # FIXME: the "time" and "epoch" needs to be fixed in strike to be in nanoseconds
        event_kwargs = rankingstat.kwargs_from_triggers(triggers, offset_vector)
        ifos = sorted(event_kwargs["snrs"])
        ifo1 = ifos.pop(0)
        snrs = event_kwargs["snrs"]
        # remove segments for ifos not having horizon distance after
        # compression, which would cause an error inside
        # local_mean_horizon_distance()
        segs_nonzero = {
            ifo: seg
            for ifo, seg in event_kwargs["segments"].items()
            if len(rankingstat.terms["P_of_tref_Dh"].horizon_history[ifo]) > 0
        }
        horizons = rankingstat.terms["P_of_tref_Dh"].local_mean_horizon_distance(
            segs_nonzero, template_id=event_kwargs["template_id"]
        )

        # generate and upload plots
        for ifo2 in ifos:
            ifo_pair = ifo1[0] + ifo2[0]
            dt_ref = event_kwargs["dt"][ifo2] - event_kwargs["dt"][ifo1]
            dphi_ref = (
                event_kwargs["dphi"][ifo2] - event_kwargs["dphi"][ifo1]
            )  # FIXME: check with Leo if phase->dphi is correct

            fig = plot_dtdphi(
                rankingstat.terms["P_of_dt_dphi"],
                ifo1,
                ifo2,
                snrs,
                horizons,
                event_dt=dt_ref,
                event_dphi=dphi_ref,
            )
            filename = "{}_dtdphi_{}.{}".format(event["gid"], ifo_pair, self.format)
            if not self.no_upload:
                upload_fig(
                    fig,
                    self.client,
                    event["gid"],
                    filename=filename,
                    log_message="{} dtdphi 2D pdf plot".format(ifo_pair),
                    tagname=("dtdphi", "background"),
                )
            if self.output_path is not None:
                filename = os.path.join(self.output_path, filename)
                self.logger.info("writing %s ...", filename)
                fig.savefig(filename)
            plt.close(fig)
            self.logger.info(
                "finished processing %s dtdphi pdf plot for %s", ifo_pair, event["gid"]
            )

    def finish(self):
        """
        upload remaining files/plots before shutting down
        """
        self.handle()


# -------------------------------------------------
#                     Main
# -------------------------------------------------


def main():
    # parse arguments
    options = parse_command_line()

    # check input options
    allowed_plots = [p.name for p in Plots]
    if not options.plot:
        # make all allowed plots by default
        options.plot = allowed_plots
    else:
        for p in options.plot:
            if p not in allowed_plots:
                raise ValueError(
                    f"Unsupported option {p} for --plot. Allowed options are: "
                    f"{allowed_plots}"
                )

    # set up logging
    log_level = logging.INFO if options.verbose else logging.WARNING
    logging.basicConfig(
        format="%(asctime)s | ll_inspiral_event_plotter : %(levelname)s : %(message)s"
    )
    logger = logging.getLogger("ll_inspiral_event_plotter")
    logger.setLevel(log_level)

    # create event plotter instance
    event_plotter = EventPlotter(
        format=options.format,
        gracedb_service_url=options.gracedb_service_url,
        kafka_server=options.kafka_server,
        logger=logger,
        max_event_time=options.max_event_time,
        max_snr=options.max_snr,
        no_upload=options.no_upload,
        output_path=options.output_path,
        plot=options.plot,
        processing_cadence=options.processing_cadence,
        ranking_stat_topic=options.ranking_stat_topic,
        request_timeout=options.request_timeout,
        tag=options.tag,
        upload_topic=options.upload_topic,
    )

    # start up
    event_plotter.start()



"""An executable to add segments table to trigger databases."""

# Copyright (C) 2024-2025 Yun-Jing Huang

from argparse import ArgumentParser

from igwn_ligolw import utils as ligolw_utils
from igwn_ligolw.utils import segments as ligolw_segments

from sgnl import sgnlio

process_name = "sgnl-add-segments"


#
# =============================================================================
#
#                                 Command Line
#
# =============================================================================
#


def parse_command_line():
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("-s", "--config-schema", help="config schema yaml file")
    parser.add_argument(
        "-i",
        "--input-database-file",
        metavar="filename",
        action="store",
        required=True,
        help="Provide the name of an input trigger database.",
    )
    parser.add_argument(
        "-o",
        "--output-database-file",
        metavar="filename",
        required=True,
        help="Provide the name of an output trigger database.",
    )
    parser.add_argument(
        "--segments-file",
        metavar="filename",
        required=True,
        help="Provide the segments file name.",
    )
    parser.add_argument(
        "--segments-name",
        metavar="name",
        required=True,
        help="Provide the name of the segments to add to the database.",
    )
    parser.add_argument("--verbose", "-v", action="store_true", help="Be verbose.")
    options = parser.parse_args()

    process_params = options.__dict__.copy()

    return options, process_params


def init_config_row(self, table, extra=None):
    out = {c["name"]: None for c in table["columns"] if not c["name"].startswith("__")}
    if extra is not None:
        out.update(extra)
    return out


#
# =============================================================================
#
#                                     Main
#
# =============================================================================
#


def main():
    #
    # Parse command line
    #

    options, process_params = parse_command_line()

    #
    # Iterate over database
    #

    indb = sgnlio.SgnlDB(
        config=options.config_schema, dbname=options.input_database_file
    )

    #
    # record our passage
    #

    indb.default_cursor.execute(
        """UPDATE process SET program = ?;""",
        (process_name,),
    )
    for name, val in process_params.items():
        indb.default_cursor.execute(
            """
        INSERT INTO process_params (param, program, value)
        VALUES (?, ?, ?);
        """,
            (name, process_name, str(val)),
        )
    #
    # Add segments
    #
    frame_segments = ligolw_segments.segmenttable_get_by_name(
        ligolw_utils.load_filename(
            options.segments_file,
            contenthandler=ligolw_segments.LIGOLWContentHandler,
        ),
        options.segments_name,
    ).coalesce()
    indb.default_cursor.execute(
        """UPDATE process SET program = ?;""",
        (process_name,),
    )
    for ifo, segs in frame_segments.items():
        for seg in segs:
            indb.default_cursor.execute(
                """
            INSERT INTO segment (start_time, end_time, ifo, name)
            VALUES (?, ?, ?, ?);
            """,
                (seg[0].ns(), seg[1].ns(), ifo, options.segments_name),
            )

    indb.to_file(options.output_database_file)



"""Tests for sgnl.bin.ll_dq"""

import sys
from unittest import mock

import pytest


@pytest.fixture(autouse=True)
def mock_dependencies():
    """Mock external dependencies and clean up after tests."""
    original_modules = {}

    modules_to_mock = [
        "sgn",
        "sgn.apps",
        "sgn.base",
        "sgnligo",
        "sgnligo.sinks",
        "sgnligo.sources",
        "sgnligo.transforms",
        "sgnts",
        "sgnts.sinks",
        "strike",
        "strike.config",
        "sgnl.psd",
        "sgnl.transforms",
    ]

    for mod in modules_to_mock:
        original_modules[mod] = sys.modules.get(mod)
        sys.modules[mod] = mock.MagicMock()

    yield

    # Restore originals
    for mod in modules_to_mock:
        if original_modules[mod] is None:
            sys.modules.pop(mod, None)
        else:
            sys.modules[mod] = original_modules[mod]

    # Clear cached imports
    sys.modules.pop("sgnl.bin.ll_dq", None)


class TestParseCommandLine:
    """Tests for parse_command_line function."""

    def test_required_args(self):
        """Test parsing with minimal required arguments."""
        from sgnl.bin import ll_dq

        with mock.patch("sys.argv", ["ll_dq"]):
            options = ll_dq.parse_command_line()
            # Check defaults
            assert options.analysis_tag == "test"
            assert options.horizon_approximant == "IMRPhenomD"
            assert options.horizon_f_min == 15.0
            assert options.horizon_f_max == 900.0
            assert options.injections is False
            assert options.search is None
            assert options.verbose is False

    def test_output_kafka_server_option(self):
        """Test --output-kafka-server option."""
        from sgnl.bin import ll_dq

        with mock.patch("sys.argv", ["ll_dq", "--output-kafka-server", "kafka:9092"]):
            options = ll_dq.parse_command_line()
            assert options.output_kafka_server == "kafka:9092"

    def test_analysis_tag_option(self):
        """Test --analysis-tag option."""
        from sgnl.bin import ll_dq

        with mock.patch("sys.argv", ["ll_dq", "--analysis-tag", "production"]):
            options = ll_dq.parse_command_line()
            assert options.analysis_tag == "production"

    def test_horizon_approximant_option(self):
        """Test --horizon-approximant option."""
        from sgnl.bin import ll_dq

        with mock.patch("sys.argv", ["ll_dq", "--horizon-approximant", "TaylorF2"]):
            options = ll_dq.parse_command_line()
            assert options.horizon_approximant == "TaylorF2"

    def test_horizon_f_min_option(self):
        """Test --horizon-f-min option."""
        from sgnl.bin import ll_dq

        with mock.patch("sys.argv", ["ll_dq", "--horizon-f-min", "20.0"]):
            options = ll_dq.parse_command_line()
            assert options.horizon_f_min == 20.0

    def test_horizon_f_max_option(self):
        """Test --horizon-f-max option."""
        from sgnl.bin import ll_dq

        with mock.patch("sys.argv", ["ll_dq", "--horizon-f-max", "1024.0"]):
            options = ll_dq.parse_command_line()
            assert options.horizon_f_max == 1024.0

    def test_injections_flag(self):
        """Test --injections flag."""
        from sgnl.bin import ll_dq

        with mock.patch("sys.argv", ["ll_dq", "--injections"]):
            options = ll_dq.parse_command_line()
            assert options.injections is True

    def test_search_ew_option(self):
        """Test --search ew option."""
        from sgnl.bin import ll_dq

        with mock.patch("sys.argv", ["ll_dq", "--search", "ew"]):
            options = ll_dq.parse_command_line()
            assert options.search == "ew"

    def test_verbose_flag(self):
        """Test --verbose flag."""
        from sgnl.bin import ll_dq

        with mock.patch("sys.argv", ["ll_dq", "--verbose"]):
            options = ll_dq.parse_command_line()
            assert options.verbose is True

    def test_verbose_short_flag(self):
        """Test -v flag."""
        from sgnl.bin import ll_dq

        with mock.patch("sys.argv", ["ll_dq", "-v"]):
            options = ll_dq.parse_command_line()
            assert options.verbose is True


class TestLlDq:
    """Tests for ll_dq function."""

    def test_ll_dq_basic(self):
        """Test ll_dq function with basic inputs."""
        from sgnl.bin import ll_dq

        # Create mock data_source_info
        mock_data_source_info = mock.MagicMock()
        mock_data_source_info.ifos = ["H1"]
        mock_data_source_info.data_source = "devshm"
        mock_data_source_info.input_sample_rate = 16384

        # Create mock condition_info
        mock_condition_info = mock.MagicMock()

        # Mock the pipeline and other components
        mock_pipeline = mock.MagicMock()
        ll_dq.Pipeline.return_value = mock_pipeline

        # Mock datasource and condition returns
        ll_dq.datasource.return_value = ({"H1": "source_link"}, None)
        ll_dq.condition.return_value = (
            {"H1": "condition_link"},
            {"H1": "spectrum_link"},
            None,
        )

        ll_dq.ll_dq(
            data_source_info=mock_data_source_info,
            condition_info=mock_condition_info,
            output_kafka_server="kafka:9092",
            analysis_tag="test",
            horizon_approximant="IMRPhenomD",
            horizon_f_min=15.0,
            horizon_f_max=900.0,
            injections=False,
            highpass_filter=10.0,
            verbose=False,
        )

        # Verify Pipeline was created and run
        ll_dq.Pipeline.assert_called_once()
        mock_pipeline.run.assert_called_once()

        # Verify datasource was called
        ll_dq.datasource.assert_called_once()

        # Verify condition was called
        ll_dq.condition.assert_called_once()

        # Verify insert was called (at least twice - once for nodes, once for links)
        assert mock_pipeline.insert.call_count >= 2

    def test_ll_dq_with_injections(self):
        """Test ll_dq function with injections enabled."""
        from sgnl.bin import ll_dq

        mock_data_source_info = mock.MagicMock()
        mock_data_source_info.ifos = ["L1"]
        mock_data_source_info.data_source = "devshm"
        mock_data_source_info.input_sample_rate = 16384

        mock_condition_info = mock.MagicMock()

        mock_pipeline = mock.MagicMock()
        ll_dq.Pipeline.return_value = mock_pipeline

        ll_dq.datasource.return_value = ({"L1": "source_link"}, None)
        ll_dq.condition.return_value = (
            {"L1": "condition_link"},
            {"L1": "spectrum_link"},
            None,
        )

        # Reset the mock to clear any previous calls
        ll_dq.KafkaSink.reset_mock()

        ll_dq.ll_dq(
            data_source_info=mock_data_source_info,
            condition_info=mock_condition_info,
            output_kafka_server="kafka:9092",
            analysis_tag="prod",
            horizon_approximant="IMRPhenomD",
            horizon_f_min=15.0,
            horizon_f_max=900.0,
            injections=True,
            highpass_filter=10.0,
            verbose=True,
        )

        # Verify KafkaSink was created with injection prefix
        kafka_sink_call = ll_dq.KafkaSink.call_args
        assert kafka_sink_call is not None
        # The prefix should contain "inj_" when injections=True
        assert "inj_" in kafka_sink_call[1]["prefix"]

    def test_ll_dq_multiple_ifos_raises(self):
        """Test ll_dq raises ValueError with multiple ifos."""
        from sgnl.bin import ll_dq

        mock_data_source_info = mock.MagicMock()
        mock_data_source_info.ifos = ["H1", "L1"]  # Multiple ifos

        mock_condition_info = mock.MagicMock()

        with pytest.raises(ValueError, match="Only supports one ifo"):
            ll_dq.ll_dq(
                data_source_info=mock_data_source_info,
                condition_info=mock_condition_info,
                output_kafka_server="kafka:9092",
                analysis_tag="test",
                horizon_approximant="IMRPhenomD",
                horizon_f_min=15.0,
                horizon_f_max=900.0,
                injections=False,
                highpass_filter=10.0,
                verbose=False,
            )


class TestMain:
    """Tests for main function."""

    def test_main_default_config(self):
        """Test main function with default config (no search)."""
        from sgnl.bin import ll_dq

        mock_data_source_info = mock.MagicMock()
        mock_data_source_info.ifos = ["H1"]
        mock_data_source_info.data_source = "devshm"
        mock_data_source_info.input_sample_rate = 16384

        mock_condition_info = mock.MagicMock()

        ll_dq.DataSourceInfo.from_options.return_value = mock_data_source_info
        ll_dq.ConditionInfo.from_options.return_value = mock_condition_info

        # Setup config mock
        mock_config = {
            "default": {"highpass_filter": 10.0},
            "ew": {"highpass_filter": 5.0},
        }
        ll_dq.get_analysis_config.return_value = mock_config

        mock_pipeline = mock.MagicMock()
        ll_dq.Pipeline.return_value = mock_pipeline

        ll_dq.datasource.return_value = ({"H1": "source_link"}, None)
        ll_dq.condition.return_value = (
            {"H1": "condition_link"},
            {"H1": "spectrum_link"},
            None,
        )

        with mock.patch("sys.argv", ["ll_dq"]):
            ll_dq.main()

        # Verify get_analysis_config was called
        ll_dq.get_analysis_config.assert_called_once()

        # Verify pipeline was run
        mock_pipeline.run.assert_called_once()

    def test_main_with_search_ew(self):
        """Test main function with ew search option."""
        from sgnl.bin import ll_dq

        mock_data_source_info = mock.MagicMock()
        mock_data_source_info.ifos = ["H1"]
        mock_data_source_info.data_source = "devshm"
        mock_data_source_info.input_sample_rate = 16384

        mock_condition_info = mock.MagicMock()

        ll_dq.DataSourceInfo.from_options.return_value = mock_data_source_info
        ll_dq.ConditionInfo.from_options.return_value = mock_condition_info

        # Setup config mock with ew-specific settings
        mock_config = {
            "default": {"highpass_filter": 10.0},
            "ew": {"highpass_filter": 5.0},
        }
        ll_dq.get_analysis_config.return_value = mock_config

        mock_pipeline = mock.MagicMock()
        ll_dq.Pipeline.return_value = mock_pipeline

        ll_dq.datasource.return_value = ({"H1": "source_link"}, None)
        ll_dq.condition.return_value = (
            {"H1": "condition_link"},
            {"H1": "spectrum_link"},
            None,
        )

        with mock.patch("sys.argv", ["ll_dq", "--search", "ew"]):
            ll_dq.main()

        # Verify pipeline was run
        mock_pipeline.run.assert_called_once()

    def test_main_with_all_options(self):
        """Test main function with all options specified."""
        from sgnl.bin import ll_dq

        mock_data_source_info = mock.MagicMock()
        mock_data_source_info.ifos = ["L1"]
        mock_data_source_info.data_source = "devshm"
        mock_data_source_info.input_sample_rate = 16384

        mock_condition_info = mock.MagicMock()

        ll_dq.DataSourceInfo.from_options.return_value = mock_data_source_info
        ll_dq.ConditionInfo.from_options.return_value = mock_condition_info

        mock_config = {
            "default": {"highpass_filter": 10.0},
            "ew": {"highpass_filter": 5.0},
        }
        ll_dq.get_analysis_config.return_value = mock_config

        mock_pipeline = mock.MagicMock()
        ll_dq.Pipeline.return_value = mock_pipeline

        ll_dq.datasource.return_value = ({"L1": "source_link"}, None)
        ll_dq.condition.return_value = (
            {"L1": "condition_link"},
            {"L1": "spectrum_link"},
            None,
        )

        with mock.patch(
            "sys.argv",
            [
                "ll_dq",
                "--output-kafka-server",
                "kafka:9092",
                "--analysis-tag",
                "production",
                "--horizon-approximant",
                "TaylorF2",
                "--horizon-f-min",
                "20.0",
                "--horizon-f-max",
                "1024.0",
                "--injections",
                "-v",
            ],
        ):
            ll_dq.main()

        # Verify pipeline was run
        mock_pipeline.run.assert_called_once()



"""A program to compute the likelhood ratios of inspiral triggers"""

# Copyright (C) 2010--2014  Kipp Cannon, Chad Hanna
# Copyright (C) 2023  Leo Tsukada


#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#


import logging
import os
import sys
from argparse import ArgumentParser

import igwn_segments as segments
from igwn_ligolw import ligolw, lsctables
from igwn_ligolw.array import use_in as array_use_in
from igwn_ligolw.param import use_in as param_use_in
from lal.utils import CacheEntry
from strike.stats import far

from sgnl import sgnlio

process_name = "sgnl-assign-likelihood"


#
# =============================================================================
#
#                                 Command Line
#
# =============================================================================
#


def parse_command_line():
    parser = ArgumentParser()
    parser.add_argument("-s", "--config-schema", help="config schema yaml file")
    parser.add_argument(
        "-i",
        "--input-database-file",
        metavar="filename",
        default=[],
        action="append",
        nargs="+",
        help="Provide the name of an input trigger database. This can be given "
        "multiple times, being separated by a space, e.g., --input-database-file "
        "filename1 filename2 filename3.",
    )
    parser.add_argument(
        "-o",
        "--output-database-file",
        metavar="filename",
        default=[],
        action="append",
        nargs="+",
        help="Provide the name of an output trigger database. This can be given "
        "multiple times, being separated by a space, e.g., --output-database-file "
        "filename1 filename2 filename3.",
    )
    parser.add_argument(
        "--input-database-cache",
        metavar="filename",
        help="Also process the files named in this LAL cache.  See "
        "lalapps_path2cache for information on how to produce a LAL cache file.",
    )
    parser.add_argument(
        "--output-database-cache",
        metavar="filename",
        help="Also write into the files named in this LAL cache.  See "
        "lalapps_path2cache for information on how to produce a LAL cache file.",
    )
    parser.add_argument(
        "-l",
        "--input-likelihood-file",
        metavar="URL",
        action="append",
        nargs="+",
        help="Set the name of the likelihood ratio data file to use. "
        "This can be given multiple times, being separated by a space, e.g., "
        "--input-likelihood-file filename1 filename2 filename3.",
    )
    parser.add_argument(
        "--input-likelihood-cache",
        metavar="filename",
        help="Also load the likelihood ratio data files listsed in this LAL "
        "cache.  See lalapps_path2cache for information on how to produce a "
        "LAL cache file.",
    )
    parser.add_argument(
        "--gstlal",
        action="store_true",
        help="Set if the likelihood ratio data file is in gstlal's format.",
    )
    parser.add_argument(
        "-t",
        "--tmp-space",
        metavar="path",
        help="Path to a directory suitable for use as a work area while "
        "manipulating the database file.  The database file will be worked on in "
        "this directory, and then moved to the final location when complete. "
        "This option is intended to improve performance when running in a "
        "networked environment, where there might be a local disk with higher "
        "bandwidth than is available to the filesystem on which the final output "
        "will reside.",
    )
    parser.add_argument(
        "--vetoes-name",
        metavar="name",
        help="Set the name of the segment lists to use as vetoes (default = do "
        "not apply vetoes).",
    )
    parser.add_argument(
        "--add-zerolag-to-background",
        action="store_true",
        help="Add zerolag events to background before populating coincident "
        "parameter PDF histograms",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force recomputation of likelihood values.",
    )
    parser.add_argument(
        "--verbose-level",
        metavar="name",
        default="INFO",
        help="Give 'DEBUG' if printing out LR components to a .log file",
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose.")
    options = parser.parse_args()

    paramdict = options.__dict__.copy()

    options.input_likelihood_files = []
    if options.input_likelihood_file is not None:
        options.input_likelihood_files += [
            url for sublist in options.input_likelihood_file for url in sublist
        ]
    if options.input_likelihood_cache is not None:
        options.input_likelihood_files += [
            CacheEntry(line).url for line in open(options.input_likelihood_cache)
        ]
    if not options.input_likelihood_files:
        raise ValueError("no likelihood URLs specified")

    if options.input_database_cache:
        options.input_database_file += [
            [CacheEntry(line).path] for line in open(options.input_database_cache)
        ]
    options.input_database_file = [
        url for sublist in options.input_database_file for url in sublist
    ]

    if options.output_database_cache:
        options.output_database_file += [
            [CacheEntry(line).path] for line in open(options.output_database_cache)
        ]
    options.output_database_file = [
        url for sublist in options.output_database_file for url in sublist
    ]

    for filename in options.output_database_file:
        if os.path.exists(filename):
            raise ValueError("output database %s already exists" % filename)

    if not len(options.input_database_file) == len(options.output_database_file):
        raise ValueError(
            "The number of each given databases are different. There must be "
            "one-to-one mapping between input and output dabases."
        )

    return options, paramdict


#
# =============================================================================
#
#                   Support Funcs for Likelihood Ratio Code
#
# =============================================================================
#


def trigger_veto_func(trigger, vetoseglists):
    # return True if event should be *vetoed*
    if trigger["ifo"] in vetoseglists:
        return trigger["time"] in vetoseglists[trigger["ifo"]]
    else:
        return False


#
# Default content handler for loading RankingStat objects from XML
# documents
#


@array_use_in
@param_use_in
@lsctables.use_in
class LIGOLWContentHandler(ligolw.LIGOLWContentHandler):
    pass


#
# =============================================================================
#
#                                     Main
#
# =============================================================================
#


def main():
    #
    # command line
    #

    options, process_params = parse_command_line()

    #
    # set up logger
    #

    logger = logging.getLogger("LR_components")
    if options.verbose_level == "DEBUG":
        logger.setLevel(logging.DEBUG)
        logging.basicConfig(
            filename="assign_likelihood_terms.log",
            filemode="w",
            format="%(name)s - %(levelname)s - %(message)s",
        )
    else:
        logger.setLevel(logging.INFO)

    #
    # load parameter distribution data
    #

    rankingstat = far.marginalize_pdf_urls(
        options.input_likelihood_files, which="LnLikelihoodRatio"
    )

    #
    # apply density estimation kernels, etc.
    # include zero-lag candidates in ranking statistic denominator if requested
    #

    rankingstat.finish(add_zerolag=options.add_zerolag_to_background)

    #
    # iterate over candidate files
    #

    failed = []
    for n, (input_database, output_database) in enumerate(
        zip(options.input_database_file, options.output_database_file), start=1
    ):

        if options.verbose:
            print("%d/%d:" % (n, len(options.input_database_file)), file=sys.stderr)
        try:
            indb = sgnlio.SgnlDB(config=options.config_schema, dbname=input_database)
        except Exception as e:
            if options.verbose:
                print(
                    "failed to load '%s': %s.  trying to continue with remaining files"
                    % (input_database, str(e)),
                    file=sys.stderr,
                )
            failed.append(input_database)
            continue

        #
        # Check if the likelihood ratio have already been populated in the input
        # database
        #

        if (
            not options.force
            and indb.default_cursor.execute(
                """SELECT EXISTS(SELECT * FROM process WHERE program = ?);""",
                (process_name,),
            ).fetchone()[0]
        ):
            if options.verbose:
                print(
                    "already processed, skipping %s" % input_database, file=sys.stderr
                )
            continue

        #
        # record our passage.
        # FIXME can we safely assume the given database always has coinc
        # information? If not, what check should there be here?
        #

        # try:
        #     coinc_def_id = lsctables.CoincDefTable.get_table(xmldoc).get_coinc_def_id(
        #         thinca.InspiralCoincDef.search,
        #         thinca.InspiralCoincDef.search_coinc_type,
        #         create_new=False,
        #     )
        # except KeyError:
        #     if options.verbose:
        #         print(
        #             "document does not contain inspiral coincidences.  skipping.",
        #             file=sys.stderr,
        #         )
        #     xmldoc.unlink()
        #     continue

        indb.default_cursor.execute(
            """UPDATE process SET program = ?;""", (process_name,)
        )
        for name, val in process_params.items():
            indb.default_cursor.execute(
                """
            INSERT INTO process_params (param, program, value)
            VALUES (?, ?, ?);
            """,
                (name, process_name, str(val)),
            )

        # FIXME : at that moment, there is no time slide. We might revisit
        # this later. It is likely that the mapping in get_events() needs to be
        # modified and nothing here should change.
        offset_vectors = {instrument: 0 for instrument in rankingstat.instruments}

        if options.vetoes_name is not None:
            # FIXME : can we get xmldoc like structure from the new db schema so
            # we can reuse ligolw_segments module?
            # vetoseglists = ligolw_segments.segmenttable_get_by_name(
            #     xmldoc, options.vetoes_name
            # ).coalesce()
            raise ValueError("vetoing feature has not implemented yet.")
        else:
            vetoseglists = segments.segmentlistdict()
        if options.verbose:
            print("done", file=sys.stderr)

        #
        # extract coinc-trigger mapping information
        #

        events = []
        for event in indb.get_events(nanosec_to_sec=True, template_duration=True):
            events += [(event["event"]["__event_id"], event["trigger"])]

        if options.verbose_level == "DEBUG":
            print(
                "Warning: dumping %d events of LR term information into a log file..."
                % len(events),
                file=sys.stderr,
            )

        #
        # assign likelihood ratio
        #

        # FIXME : calc_likelihood.assign_likelihood_ratios_xml is no longer
        # used. this might harm the computation efficiency (?), in which case we
        # might want to revisit this.
        for event_id, triggers in events:
            if any(trigger_veto_func(trigger, vetoseglists) for trigger in triggers):
                continue
            indb.default_cursor.execute(
                """
            UPDATE event SET likelihood=? WHERE __event_id=?;
                                        """,
                (
                    float(rankingstat.ln_lr_from_triggers(triggers, offset_vectors)),
                    event_id,
                ),
            )

        #
        # close out process metadata.
        #

        # FIXME : figure out how to check in these information for the new CBC db
        #         schema using stillsuit package
        # process.set_end_time_now()
        # connection.cursor().execute(
        #     "UPDATE process SET end_time = ? WHERE process_id == ?",
        #     (process.end_time, process.process_id),
        # )

        if options.verbose:
            print(
                "likelihood assignment complete for %s" % input_database,
                file=sys.stderr,
            )
            print("Writing to %s" % output_database, file=sys.stderr)
        indb.to_file(output_database)

    #
    # crash if any input files were broken, otherwise exit the program
    #

    if failed:
        raise ValueError(
            "%s could not be processed" % ", ".join("'%s'" % url for url in failed)
        )
    elif options.verbose:
        print("Done", file=sys.stderr)


if __name__ == "__main__":
    main()



"""Populate zerolag pdf and apply the extinction model"""

# Copyright (C) 2023 Prathamesh Joshi
# Copyright (C) 2024 Leo Tsukada


#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#


import sys
from argparse import ArgumentParser

import numpy
import scipy.optimize
import stillsuit
from strike.stats import far

#
# =============================================================================
#
#                                 Command Line
#
# =============================================================================
#


def parse_command_line():
    parser = ArgumentParser()
    parser.add_argument("-s", "--config-schema", help="config schema yaml file")
    parser.add_argument(
        "--input-database-file",
        metavar="filename",
        help="Input trigger database from which to add zerolag. Optional",
    )
    parser.add_argument(
        "--input-rankingstatpdf-file",
        metavar="filename",
        action="append",
        nargs="+",
        default=[],
        help="Set the name of the input RANK_STAT_PDF file. This can be given "
        "multiple times, being separated by a space, e.g., --input-rankingstatpdf-file "
        "filename1 filename2 filename3.",
    )
    parser.add_argument(
        "--output-rankingstatpdf-file",
        metavar="filename",
        help="Set the name of the output RANK_STAT_PDF file. If none is provided, the "
        "input file will be overwritten",
    )
    parser.add_argument(
        "--reset-zerolag",
        action="store_true",
        default=False,
        help="Reset the zerolag to zero after performing the extinction",
    )
    parser.add_argument(
        "--initial-exponent",
        default=1.0,
        help="Set the initial guess for the exponent during extinction fitting",
    )
    parser.add_argument("--verbose", "-v", action="store_true", help="Be verbose.")
    options = parser.parse_args()

    options.input_rankingstatpdf_file = [
        url for sublist in options.input_rankingstatpdf_file for url in sublist
    ]

    if options.output_rankingstatpdf_file is None:
        options.output_rankingstatpdf_file = options.input_rankingstatpdf_file[0]
    return options


#
# =============================================================================
#
#                                     Main
#
# =============================================================================
#


def main():

    options = parse_command_line()

    # load and marginalize the rankingstatpdf
    rankingstatpdf = far.marginalize_pdf_urls(
        options.input_rankingstatpdf_file, which="RankingStatPDF"
    )

    if options.input_database_file is not None:
        db = stillsuit.StillSuit(
            config=options.config_schema, dbname=options.input_database_file
        )
        rankingstatpdf.collect_zero_lag_rates(db)

    bg = rankingstatpdf.noise_lr_lnpdf.array
    fg = rankingstatpdf.zero_lag_lr_lnpdf.array

    # zero out the beginning bins of each because they are notoriously bad and should
    # just be ignored
    bg[:10] = 0.0
    fg[:10] = 0.0

    # fitting is done between ix_min and ix_max
    fg_ccdf = numpy.cumsum(fg[::-1])[::-1]
    ix_min = (fg_ccdf < fg_ccdf[0] / 2.0).argmax()
    ix_max = (fg_ccdf < fg_ccdf[0] / 100.0).argmax()

    bgtotal = bg[ix_min : ix_max + 1].sum()
    bg_ccdf = numpy.cumsum(bg[::-1])[::-1] / bgtotal

    # define a function for the extincted bg for scipy.optimize.curve_fit to call
    def bg_ccdf_extinct_func(idx, c, A):
        return numpy.log(A * bgtotal / c) + numpy.log1p(
            -numpy.exp(-1 * bg_ccdf[idx] * c)
        )

    # find the best fit c for extinction
    c = scipy.optimize.curve_fit(
        bg_ccdf_extinct_func,
        range(ix_min, ix_max + 1),
        numpy.log(fg_ccdf[ix_min : ix_max + 1]),
        p0=[float(options.initial_exponent), 1],
        bounds=[0, numpy.inf],
        sigma=numpy.sqrt(1 / (bg_ccdf[ix_min : ix_max + 1] * bgtotal)),
    )  # , maxfev = 5000)
    print(
        f"Best value of c is {c[0][0]}, A is {c[0][1]} with covariance {c[1]}",
        file=sys.stderr,
    )

    # calculate the extincted PDF
    bg_pdf_extinct = c[0][1] * bg * numpy.exp(-1 * bg_ccdf * c[0][0])

    rankingstatpdf.noise_lr_lnpdf.array = bg_pdf_extinct
    rankingstatpdf.noise_lr_lnpdf.normalize()

    if options.reset_zerolag:
        for i in range(len(rankingstatpdf.zero_lag_lr_lnpdf.array)):
            rankingstatpdf.zero_lag_lr_lnpdf.array[i] = 0.0

    rankingstatpdf.save(
        options.output_rankingstatpdf_file,
        process_name="sgnl-extinct-bin",
        process_params=options.__dict__,
    )


if __name__ == "__main__":
    main()



"""An executable to measure the power spectral density of time-series data."""

# Copyright (C) 2010 Kipp Cannon, Chad Hanna, Leo Singer
# Copyright (C) 2024 Anushka Doke, Yun-Jing Huang, Ryan Magee, Shio Sakon

from __future__ import annotations

import os
from argparse import ArgumentParser

from sgn.apps import Pipeline
from sgnligo.sources import DataSourceInfo, datasource
from sgnligo.transforms import ConditionInfo, condition
from sgnts.sinks import DumpSeriesSink, NullSeriesSink
from strike.config import get_analysis_config

from sgnl.sinks import PSDSink

default_config = get_analysis_config()["default"]


def parse_command_line():
    parser = ArgumentParser(description=__doc__)

    DataSourceInfo.append_options(parser)
    ConditionInfo.append_options(parser)

    # add our own options

    parser.add_argument(
        "--search",
        type=str,
        default=None,
        choices=["ew", None],
        help="Set the search, if you want search-specific changes to be implemented "
        "while data whitening. Allowed choices: ['ew', None].",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Be verbose (optional)."
    )
    parser.add_argument(
        "--output-name",
        metavar="filename",
        required=True,
        help="Set the output filename. If no filename is given, the output is dumped"
        " to stdout.",
    )
    parser.add_argument(
        "--whitened-data-output-name",
        metavar="filename",
        help="Set the output filename for dumping whitened data into",
    )

    options = parser.parse_args()

    return options


def reference_psd(
    data_source_info: DataSourceInfo,
    condition_info: ConditionInfo,
    highpass_filter=default_config["highpass_filter"],
    output_name=None,
    whitened_data_output_name=None,
    verbose=None,
):

    pipeline = Pipeline()

    # Create data source
    source_out_links, _ = datasource(
        pipeline=pipeline,
        info=data_source_info,
    )

    #
    #          -----------
    #         | DevShmSource |
    #          -----------
    #         /
    #     H1 /
    #   ------------
    #  |  Resampler |
    #   ------------
    #       |
    #   ------------  hoft ----------
    #  |  Whiten    | --- | NullSink |
    #   ------------       ----------
    #          |psd
    #   ------------
    #  |  PSDSink   |
    #   ------------
    #

    ifos = data_source_info.ifos

    condition_out_links, spectrum_out_links, _ = condition(
        pipeline=pipeline,
        condition_info=condition_info,
        ifos=ifos,
        data_source=data_source_info.data_source,
        input_sample_rate=data_source_info.input_sample_rate,
        input_links=source_out_links,
        highpass_filter=highpass_filter,
    )

    for ifo in ifos:
        if whitened_data_output_name:
            outname = os.path.join(
                os.path.dirname(whitened_data_output_name),
                ifo + "_" + os.path.basename(whitened_data_output_name),
            )
            pipeline.insert(
                DumpSeriesSink(
                    name=ifo + "HoftSink",
                    sink_pad_names=("hoft",),
                    fname=outname,
                    verbose=verbose,
                ),
                link_map={
                    ifo + "HoftSink:snk:hoft": condition_out_links[ifo],
                },
            )
        else:
            pipeline.insert(
                NullSeriesSink(
                    name=ifo + "HoftSink",
                    sink_pad_names=("hoft",),
                    verbose=verbose,
                ),
                link_map={
                    ifo + "HoftSink:snk:hoft": condition_out_links[ifo],
                },
            )
    pipeline.insert(
        PSDSink(
            fname=output_name,
            name="PSDSink",
            sink_pad_names=(tuple(ifo for ifo in ifos)),
        ),
        link_map={"PSDSink:snk:" + ifo: spectrum_out_links[ifo] for ifo in ifos},
    )

    pipeline.run()


def main():
    # parse arguments
    options = parse_command_line()

    data_source_info = DataSourceInfo.from_options(options)
    condition_info = ConditionInfo.from_options(options)

    config = get_analysis_config()
    config = config[options.search] if options.search else config["default"]

    reference_psd(
        data_source_info=data_source_info,
        condition_info=condition_info,
        highpass_filter=config["highpass_filter"],
        output_name=options.output_name,
        whitened_data_output_name=options.whitened_data_output_name,
        verbose=options.verbose,
    )


if __name__ == "__main__":
    main()



"""A program to compute the LLOID filtering costs and memory usage of an SVD bank as
produced by sgnl-inspiral-svd-bank.
"""

# Copyright (C) 2014  Chad Hanna, Miguel Fernandez
# Copyright (C) 2025  Yun-Jing Huang

from argparse import ArgumentParser

import h5py
import numpy

from sgnl import svd_bank


def parse_command_line():
    parser = ArgumentParser()

    parser.add_argument(
        "--svd-bank",
        metavar="filename",
        nargs="+",
        required=True,
        help="Set the name of the LIGO light-weight XML file from which to load the "
        "svd bank for a given instrument.  To analyze multiple instruments, --svd-bank "
        "can be called multiple times for svd banks corresponding to different "
        "instruments.  If --data-source is lvshm or framexmit, then only svd banks "
        "corresponding to a single bin must be given. If given multiple times, the "
        "banks will be processed one-by-one, in order.  At least one svd bank for at "
        "least 2 detectors is required, but see also --svd-bank-cache.",
    )
    parser.add_argument(
        "--output-path",
        metavar="filename",
        required=True,
        help="Output filename.",
    )
    parser.add_argument(
        "--memory",
        action="store_true",
        help="Calculate memory usage.",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Be verbose.",
    )
    options = parser.parse_args()

    return options


def flopulator(r, rT, UT, MT, NT, verbose=False):
    # Convolution of a shape = (upfactor, 17) filter requires a multiply-add
    # per sample point of data for each sample of the filter for each physical
    # template. Don't upsample the highest sample rate
    kernel = 17
    upfactor = numpy.array([r[i + 1] / ri for i, ri in enumerate(r[:-1])])
    resample = (r[:-1] * upfactor * kernel * 2 * MT).sum()

    # Convolution of a NT sample filter requires a multiply-add per sample
    # point of data for each sample of the filter for each svd template
    filter = (NT * rT * UT * 2).sum()

    reconstruct = (MT * UT * rT * 2).sum()

    add = (rT[:-1] * MT).sum()

    # get FLOPs per *complex* template (note the divide by 2)
    if verbose:
        print("--->\tMFLOPS from resampling: ", resample / 1000.0**2)
        print("--->\tMFLOPS from filtering: ", filter / 1000.0**2)
        print("--->\tMFLOPS from reconstruction: ", reconstruct / 1000.0**2)
        print("--->\tMFLOPS from addition: ", add / 1000.0**2)

    MFLOPs = (
        resample / 1000.0**2
        + filter / 1000.0**2
        + reconstruct / 1000.0**2
        + add / 1000.0**2
    )

    if verbose:
        print("--->\tTotal MFLOPS: ", MFLOPs)
        print("--->\tMFLOPS per complex template: ", MFLOPs / (MT / 2.0))
        print(
            "--->\tRatio number SVD filters to number real input templates: ",
            UT.sum() / MT,
        )

    return MFLOPs


def memulator(r, rT, UT, MT, NT, verbose=False):
    # FIXME: I don't know if these calculations are correct

    # Convolution of a shape = (upfactor, 17) filter requires a multiply-add
    # per sample point of data for each sample of the filter for each physical
    # template. Don't upsample the highest sample rate
    # inputs: (upfactor, 17), (MT, 16 + rate)
    # output: (MT, upfactor * rate)
    kernel = 17
    upfactor = numpy.array([r[i + 1] / ri for i, ri in enumerate(r[:-1])])
    resample = (upfactor * kernel + MT * (16 + r[:-1]) + MT * (upfactor * r[:-1])).sum()

    # Convolution of a NT sample filter requires a multiply-add per sample
    # point of data for each sample of the filter for each svd template
    # inputs: (UT, NT), (rate + NT -1,)
    # outputs: (UT, rate)
    filter = (UT * NT + rT + NT - 1 + UT * rT).sum()

    # inputs: (MT, UT), (UT, rate)
    # outputs: (MT, rate)
    reconstruct = (MT * UT + UT * rT + MT * rT).sum()

    # inputs: (MT, rate), (MT, rate)
    # outputs: (MT, rate)
    add = (3 * MT * rT[:-1]).sum()

    # get FLOPs per *complex* template (note the divide by 2)
    if verbose:
        print("--->\tMB/s from resampling: ", resample * 4 / 1000.0**2)
        print("--->\tMB/s from filtering: ", filter * 4 / 1000.0**2)
        print("--->\tMB/s from reconstruction: ", reconstruct * 4 / 1000.0**2)
        print("--->\tMB/s from addition: ", add * 4 / 1000.0**2)

    MB = (
        resample * 4 / 1000.0**2
        + filter * 4 / 1000.0**2
        + reconstruct * 4 / 1000.0**2
        + add * 4 / 1000.0**2
    )
    if verbose:
        print("--->\tTotal MB/s: ", MB)
        print("--->\tMB/s per complex template: ", MB / (MT / 2.0))
        print(
            "--->\tRatio number SVD filters to number real input templates: ",
            UT.sum() / MT,
        )

    return MB


def main():
    options = parse_command_line()
    hf = h5py.File(options.output_path, "a")
    for f in options.svd_bank:
        banks = svd_bank.read_banks(f, svd_bank.DefaultContentHandler)
        svdbin = banks[0].bank_id.split("_")[0]
        hf.create_group(f"{svdbin}")
        totalMFLOPS = 0
        totalMB = 0
        for i, bank in enumerate(banks):
            g = hf.create_group(f"{svdbin}/subbank-{i}")
            rT = [f.rate for f in bank.bank_fragments]
            r = numpy.array(sorted(list(set(rT))))
            rT = numpy.array(rT)

            UT = numpy.array([f.mix_matrix.shape[0] for f in bank.bank_fragments])
            MT = [f.mix_matrix.shape[1] for f in bank.bank_fragments][0]
            NT = numpy.array([(f.end - f.start) * f.rate for f in bank.bank_fragments])

            if options.verbose:
                print("\nSUB BANK %d" % i)
                print("--->\tUnique sampling rates: ", r)
                print("--->\tSampling rate for a given time slice: ", rT)
                print("--->\tTotal SVD filters for a given time slice: ", UT)
                print("--->\tNumber of SVD filter samples: ", NT)
                print(
                    "--->\tTotal real templates (e.g. twice number of complex "
                    " templates): ",
                    MT,
                )

            g.create_dataset("sample-rates", data=rT)
            g.create_dataset("num-svd-filters", data=UT)
            g.create_dataset("num-real-templates", data=MT)

            totalMFLOPS += flopulator(r, rT, UT, MT, NT, options.verbose)
            if options.memory:
                totalMB += memulator(r, rT, UT, MT, NT, options.verbose)

        g.create_dataset("total-mflops", data=totalMFLOPS)
        print("SVD BIN: %s | Total MFLOPs %f" % (svdbin, totalMFLOPS))
        if options.memory:
            g.create_dataset("total-mbs", data=totalMB)
            print("SVD BIN: %s | Total MB/s %f" % (svdbin, totalMB))
    hf.close()



"""Tests for Itacacac transform's usage of EventFrame/EventBuffer API"""

import pytest
import torch
from sgnts.base import EventBuffer

from sgnl.transforms.itacacac import Itacacac, index_select, light_travel_time


def test_itacacac_output_events_structure():
    """Test that output_events returns EventBuffer with correct dictionary structure"""
    # Create a minimal Itacacac instance
    itac = Itacacac(
        name="test_itacacac",
        sink_pad_names=["H1"],
        sample_rate=1024,
        trigger_finding_duration=1.0,
        snr_min=4.0,
        autocorrelation_banks={"H1": torch.zeros(1, 1, 10, dtype=torch.complex64)},
        autocorrelation_length_mask=None,
        autocorrelation_lengths={0: 10},
        template_ids=torch.tensor([[0]]),
        bankids_map={0: [0]},
        end_time_delta=torch.tensor([0.0]),
        template_durations=torch.tensor([[1.0]]),
        device="cpu",
        coincidence_threshold=0.0,
        min_instruments_candidates=1,
        strike_pad=None,
        stillsuit_pad="stillsuit",
    )

    # Mock clustered_coinc data
    clustered_coinc = {
        "sngls": {
            "H1": {
                "time": torch.tensor([1000000000]),
                "shifted_time": torch.tensor([1000000]),
                "snr": torch.tensor([5.0]),
                "chisq": torch.tensor([1.0]),
                "phase": torch.tensor([0.0]),
            }
        },
        "clustered_template_ids": torch.tensor([0]),
        "clustered_template_durations": torch.tensor([1.0]),
        "clustered_ifo_combs": torch.tensor([1]),
        "clustered_snr": torch.tensor([5.0]),
        "clustered_bankids": [0],
        "snr_ts_snippet_clustered": {"H1": torch.zeros(1, 2, 10)},
        "snr_ts_clustered": {"H1": torch.zeros(1, 2, 100)},
    }

    ts, te = 1000000000, 1001000000
    result = itac.output_events(clustered_coinc, ts, te)

    # Test that result is an EventBuffer
    assert isinstance(result, EventBuffer)

    # Test that data is a list with one dictionary
    assert isinstance(result.data, list)
    assert len(result.data) == 1
    assert isinstance(result.data[0], dict)

    # Test that dictionary has expected keys
    expected_keys = {"event", "trigger", "snr_ts", "max_snr_histories"}
    assert set(result.data[0].keys()) == expected_keys

    # Test that event data is a list
    assert isinstance(result.data[0]["event"], list)
    assert len(result.data[0]["event"]) == 1

    # Test that trigger data is a list
    assert isinstance(result.data[0]["trigger"], list)
    assert len(result.data[0]["trigger"]) == 1


def test_itacacac_output_background_structure():
    """Test that output_background returns EventBuffer with correct structure"""
    # Create a minimal Itacacac instance with strike pad
    itac = Itacacac(
        name="test_itacacac",
        sink_pad_names=["H1"],
        sample_rate=1024,
        trigger_finding_duration=1.0,
        snr_min=4.0,
        autocorrelation_banks={"H1": torch.zeros(1, 1, 10, dtype=torch.complex64)},
        autocorrelation_length_mask=None,
        autocorrelation_lengths={0: 10},
        template_ids=torch.tensor([[0]]),
        bankids_map={0: [0]},
        end_time_delta=torch.tensor([0.0]),
        template_durations=torch.tensor([[1.0]]),
        device="cpu",
        coincidence_threshold=0.0,
        min_instruments_candidates=1,
        strike_pad="strike",
        stillsuit_pad="stillsuit",
    )

    # Mock triggers data
    triggers = {
        "snrs": {"H1": torch.tensor([[5.0]])},
        "chisqs": {"H1": torch.tensor([[1.0]])},
    }
    single_background_masks = {"H1": torch.tensor([[True]])}

    ts, te = 1000000000, 1001000000
    result = itac.output_background(triggers, single_background_masks, ts, te)

    # Test that result is an EventBuffer
    assert isinstance(result, EventBuffer)

    # Test that data is a list with one dictionary
    assert isinstance(result.data, list)
    assert len(result.data) == 1
    assert isinstance(result.data[0], dict)

    # Test that dictionary has expected keys
    expected_keys = {"trigger_rates", "background"}
    assert set(result.data[0].keys()) == expected_keys


def test_index_select():
    """Test the index_select helper function."""
    tensor = torch.tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    index = torch.tensor([0, 2, 1])
    result = index_select(tensor, dim=1, index=index)
    expected = torch.tensor([1, 6, 8])
    assert torch.equal(result, expected)


def test_light_travel_time():
    """Test light_travel_time function."""
    # Test with H1 and L1
    dt = light_travel_time("H1", "L1")
    assert dt > 0
    assert dt < 0.02  # Should be about 10ms


def test_itacacac_with_autocorrelation_length_mask():
    """Test Itacacac with autocorrelation_length_mask set."""
    mask = {"H1": torch.ones(1, 1, 10)}
    itac = Itacacac(
        name="test_itacacac",
        sink_pad_names=["H1"],
        sample_rate=1024,
        trigger_finding_duration=1.0,
        snr_min=4.0,
        autocorrelation_banks={"H1": torch.zeros(1, 1, 10, dtype=torch.complex64)},
        autocorrelation_length_mask=mask,
        autocorrelation_lengths={0: 10},
        template_ids=torch.tensor([[0]]),
        bankids_map={0: [0]},
        end_time_delta=torch.tensor([0.0]),
        template_durations=torch.tensor([[1.0]]),
        device="cpu",
        coincidence_threshold=0.0,
        min_instruments_candidates=1,
        strike_pad=None,
        stillsuit_pad="stillsuit",
    )
    # Verify autocorrelation_norms was computed with mask
    assert "H1" in itac.autocorrelation_norms


def test_make_coincs_single_ifo_min_instruments_2():
    """Test make_coincs with single ifo and min_instruments_candidates=2."""
    itac = Itacacac(
        name="test_itacacac",
        sink_pad_names=["H1"],
        sample_rate=1024,
        trigger_finding_duration=1.0,
        snr_min=4.0,
        autocorrelation_banks={"H1": torch.zeros(1, 2, 10, dtype=torch.complex64)},
        autocorrelation_length_mask=None,
        autocorrelation_lengths={0: 10},
        template_ids=torch.tensor([[0, 1]]),
        bankids_map={0: [0]},
        end_time_delta=torch.tensor([0.0]),
        template_durations=torch.tensor([[1.0, 1.0]]),
        device="cpu",
        coincidence_threshold=0.0,
        min_instruments_candidates=2,
        strike_pad=None,
        stillsuit_pad="stillsuit",
    )

    triggers = {
        "snrs": {"H1": torch.tensor([[5.0, 3.0]])},
        "peak_locations": {"H1": torch.tensor([[100, 101]])},
        "chisqs": {"H1": torch.tensor([[1.0, 1.0]])},
    }

    ifo_combs, network_snr, single_masks, noevents_mask = itac.make_coincs(triggers)

    # With min_instruments_candidates=2, single ifo should not produce events
    assert ifo_combs is None or torch.all(noevents_mask)


def test_make_coincs_nifo_greater_than_3():
    """Test make_coincs raises ValueError for nifo > 3."""
    itac = Itacacac(
        name="test_itacacac",
        sink_pad_names=["H1", "L1", "V1", "K1"],
        sample_rate=1024,
        trigger_finding_duration=1.0,
        snr_min=4.0,
        autocorrelation_banks={
            "H1": torch.zeros(1, 2, 10, dtype=torch.complex64),
            "L1": torch.zeros(1, 2, 10, dtype=torch.complex64),
            "V1": torch.zeros(1, 2, 10, dtype=torch.complex64),
            "K1": torch.zeros(1, 2, 10, dtype=torch.complex64),
        },
        autocorrelation_length_mask=None,
        autocorrelation_lengths={0: 10},
        template_ids=torch.tensor([[0, 1]]),
        bankids_map={0: [0]},
        end_time_delta=torch.tensor([0.0]),
        template_durations=torch.tensor([[1.0, 1.0]]),
        device="cpu",
        coincidence_threshold=0.0,
        min_instruments_candidates=1,
        strike_pad=None,
        stillsuit_pad="stillsuit",
    )

    triggers = {
        "snrs": {
            "H1": torch.tensor([[5.0, 3.0]]),
            "L1": torch.tensor([[5.0, 3.0]]),
            "V1": torch.tensor([[5.0, 3.0]]),
            "K1": torch.tensor([[5.0, 3.0]]),
        },
        "peak_locations": {
            "H1": torch.tensor([[100, 101]]),
            "L1": torch.tensor([[100, 101]]),
            "V1": torch.tensor([[100, 101]]),
            "K1": torch.tensor([[100, 101]]),
        },
    }

    with pytest.raises(ValueError, match="nifo > 3 is not implemented"):
        itac.make_coincs(triggers)


def test_coinc2_min_instruments_2():
    """Test coinc2 with min_instruments_candidates=2."""
    itac = Itacacac(
        name="test_itacacac",
        sink_pad_names=["H1", "L1"],
        sample_rate=1024,
        trigger_finding_duration=1.0,
        snr_min=4.0,
        autocorrelation_banks={
            "H1": torch.zeros(1, 2, 10, dtype=torch.complex64),
            "L1": torch.zeros(1, 2, 10, dtype=torch.complex64),
        },
        autocorrelation_length_mask=None,
        autocorrelation_lengths={0: 10},
        template_ids=torch.tensor([[0, 1]]),
        bankids_map={0: [0]},
        end_time_delta=torch.tensor([0.0]),
        template_durations=torch.tensor([[1.0, 1.0]]),
        device="cpu",
        coincidence_threshold=0.0,
        min_instruments_candidates=2,
        strike_pad=None,
        stillsuit_pad="stillsuit",
    )
    itac.rate = 1024

    snrs = [torch.tensor([[5.0, 3.0]]), torch.tensor([[3.0, 5.0]])]
    times = [torch.tensor([[100, 101]]), torch.tensor([[200, 201]])]

    result = itac.coinc2(snrs, times, ["H1", "L1"])

    # With min_instruments_candidates=2, single_mask should be None
    assert result[1] is None  # single_mask1
    assert result[2] is None  # single_mask2


def test_coinc3_min_instruments_2():
    """Test coinc3 with min_instruments_candidates=2."""
    itac = Itacacac(
        name="test_itacacac",
        sink_pad_names=["H1", "L1", "V1"],
        sample_rate=1024,
        trigger_finding_duration=1.0,
        snr_min=4.0,
        autocorrelation_banks={
            "H1": torch.zeros(1, 2, 10, dtype=torch.complex64),
            "L1": torch.zeros(1, 2, 10, dtype=torch.complex64),
            "V1": torch.zeros(1, 2, 10, dtype=torch.complex64),
        },
        autocorrelation_length_mask=None,
        autocorrelation_lengths={0: 10},
        template_ids=torch.tensor([[0, 1]]),
        bankids_map={0: [0]},
        end_time_delta=torch.tensor([0.0]),
        template_durations=torch.tensor([[1.0, 1.0]]),
        device="cpu",
        coincidence_threshold=0.0,
        min_instruments_candidates=2,
        strike_pad=None,
        stillsuit_pad="stillsuit",
    )
    itac.rate = 1024

    triggers = {
        "snrs": {
            "H1": torch.tensor([[5.0, 3.0]]),
            "L1": torch.tensor([[5.0, 3.0]]),
            "V1": torch.tensor([[5.0, 3.0]]),
        },
        "peak_locations": {
            "H1": torch.tensor([[100, 101]]),
            "L1": torch.tensor([[200, 201]]),
            "V1": torch.tensor([[300, 301]]),
        },
    }

    result = itac.coinc3(triggers)

    # With min_instruments_candidates=2, single masks should be None
    assert result[4] is None  # single_mask1
    assert result[5] is None  # single_mask2
    assert result[6] is None  # single_mask3


def test_coinc3_all_triggers_to_background():
    """Test coinc3 with all_triggers_to_background=True."""
    itac = Itacacac(
        name="test_itacacac",
        sink_pad_names=["H1", "L1", "V1"],
        sample_rate=1024,
        trigger_finding_duration=1.0,
        snr_min=4.0,
        autocorrelation_banks={
            "H1": torch.zeros(1, 2, 10, dtype=torch.complex64),
            "L1": torch.zeros(1, 2, 10, dtype=torch.complex64),
            "V1": torch.zeros(1, 2, 10, dtype=torch.complex64),
        },
        autocorrelation_length_mask=None,
        autocorrelation_lengths={0: 10},
        template_ids=torch.tensor([[0, 1]]),
        bankids_map={0: [0]},
        end_time_delta=torch.tensor([0.0]),
        template_durations=torch.tensor([[1.0, 1.0]]),
        device="cpu",
        coincidence_threshold=0.0,
        min_instruments_candidates=1,
        all_triggers_to_background=True,
        strike_pad=None,
        stillsuit_pad="stillsuit",
    )
    itac.rate = 1024

    triggers = {
        "snrs": {
            "H1": torch.tensor([[5.0, 3.0]]),
            "L1": torch.tensor([[5.0, 3.0]]),
            "V1": torch.tensor([[5.0, 3.0]]),
        },
        "peak_locations": {
            "H1": torch.tensor([[100, 101]]),
            "L1": torch.tensor([[200, 201]]),
            "V1": torch.tensor([[300, 301]]),
        },
    }

    result = itac.coinc3(triggers)

    # With all_triggers_to_background=True, background masks should include all
    # triggers above snr_min
    assert result[7] is not None  # single_background_mask1
    assert result[8] is not None  # single_background_mask2
    assert result[9] is not None  # single_background_mask3


def test_make_coincs_two_ifo():
    """Test make_coincs with 2 ifos."""
    itac = Itacacac(
        name="test_itacacac",
        sink_pad_names=["H1", "L1"],
        sample_rate=1024,
        trigger_finding_duration=1.0,
        snr_min=4.0,
        autocorrelation_banks={
            "H1": torch.zeros(1, 2, 10, dtype=torch.complex64),
            "L1": torch.zeros(1, 2, 10, dtype=torch.complex64),
        },
        autocorrelation_length_mask=None,
        autocorrelation_lengths={0: 10},
        template_ids=torch.tensor([[0, 1]]),
        bankids_map={0: [0]},
        end_time_delta=torch.tensor([0.0]),
        template_durations=torch.tensor([[1.0, 1.0]]),
        device="cpu",
        coincidence_threshold=0.0,
        min_instruments_candidates=1,
        strike_pad=None,
        stillsuit_pad="stillsuit",
    )
    itac.rate = 1024

    triggers = {
        "snrs": {
            "H1": torch.tensor([[5.0, 3.0]]),
            "L1": torch.tensor([[5.0, 3.0]]),
        },
        "peak_locations": {
            "H1": torch.tensor([[100, 101]]),
            "L1": torch.tensor([[100, 101]]),
        },
    }

    ifo_combs, network_snr, single_masks, noevents_mask = itac.make_coincs(triggers)

    assert ifo_combs is not None
    assert network_snr is not None


def test_make_coincs_two_ifo_all_triggers_to_background():
    """Test make_coincs with 2 ifos and all_triggers_to_background."""
    itac = Itacacac(
        name="test_itacacac",
        sink_pad_names=["H1", "L1"],
        sample_rate=1024,
        trigger_finding_duration=1.0,
        snr_min=4.0,
        autocorrelation_banks={
            "H1": torch.zeros(1, 2, 10, dtype=torch.complex64),
            "L1": torch.zeros(1, 2, 10, dtype=torch.complex64),
        },
        autocorrelation_length_mask=None,
        autocorrelation_lengths={0: 10},
        template_ids=torch.tensor([[0, 1]]),
        bankids_map={0: [0]},
        end_time_delta=torch.tensor([0.0]),
        template_durations=torch.tensor([[1.0, 1.0]]),
        device="cpu",
        coincidence_threshold=0.0,
        min_instruments_candidates=1,
        all_triggers_to_background=True,
        strike_pad=None,
        stillsuit_pad="stillsuit",
    )
    itac.rate = 1024

    triggers = {
        "snrs": {
            "H1": torch.tensor([[5.0, 3.0]]),
            "L1": torch.tensor([[5.0, 3.0]]),
        },
        "peak_locations": {
            "H1": torch.tensor([[100, 101]]),
            "L1": torch.tensor([[100, 101]]),
        },
    }

    ifo_combs, network_snr, single_masks, noevents_mask = itac.make_coincs(triggers)

    # Check single_masks includes both ifos
    assert "H1" in single_masks
    assert "L1" in single_masks


def test_make_coincs_single_ifo_all_triggers_to_background():
    """Test make_coincs with single ifo and all_triggers_to_background."""
    itac = Itacacac(
        name="test_itacacac",
        sink_pad_names=["H1"],
        sample_rate=1024,
        trigger_finding_duration=1.0,
        snr_min=4.0,
        autocorrelation_banks={"H1": torch.zeros(1, 2, 10, dtype=torch.complex64)},
        autocorrelation_length_mask=None,
        autocorrelation_lengths={0: 10},
        template_ids=torch.tensor([[0, 1]]),
        bankids_map={0: [0]},
        end_time_delta=torch.tensor([0.0]),
        template_durations=torch.tensor([[1.0, 1.0]]),
        device="cpu",
        coincidence_threshold=0.0,
        min_instruments_candidates=1,
        all_triggers_to_background=True,
        strike_pad=None,
        stillsuit_pad="stillsuit",
    )

    triggers = {
        "snrs": {"H1": torch.tensor([[5.0, 3.0]])},
        "peak_locations": {"H1": torch.tensor([[100, 101]])},
    }

    ifo_combs, network_snr, single_masks, noevents_mask = itac.make_coincs(triggers)

    # Check single_masks includes H1
    assert "H1" in single_masks


def test_output_events_online_with_padding():
    """Test output_events in online mode with different autocorr lengths."""
    itac = Itacacac(
        name="test_itacacac",
        sink_pad_names=["H1"],
        sample_rate=1024,
        trigger_finding_duration=1.0,
        snr_min=4.0,
        autocorrelation_banks={"H1": torch.zeros(1, 1, 10, dtype=torch.complex64)},
        autocorrelation_length_mask=None,
        autocorrelation_lengths={0: 6},  # Dict mapping bankid to length
        template_ids=torch.tensor([[0]]),
        bankids_map={0: [0]},
        end_time_delta=torch.tensor([0.0]),
        template_durations=torch.tensor([[1.0]]),
        device="cpu",
        coincidence_threshold=0.0,
        min_instruments_candidates=1,
        strike_pad=None,
        stillsuit_pad="stillsuit",
        is_online=True,
    )

    # Create clustered_coinc with snr_ts_snippet that needs padding trimmed
    clustered_coinc = {
        "sngls": {
            "H1": {
                "time": torch.tensor([1000000000]),
                "shifted_time": torch.tensor([1000000]),
                "snr": torch.tensor([5.0]),
                "chisq": torch.tensor([1.0]),
                "phase": torch.tensor([0.0]),
            }
        },
        "clustered_template_ids": torch.tensor([0]),
        "clustered_template_durations": torch.tensor([1.0]),
        "clustered_ifo_combs": torch.tensor([1]),
        "clustered_snr": torch.tensor([5.0]),
        "clustered_bankids": [0],
        "snr_ts_snippet_clustered": {
            "H1": torch.zeros(1, 2, 10)
        },  # 10 samples, will be trimmed to 6
        "snr_ts_clustered": {"H1": torch.zeros(1, 2, 100)},
    }

    ts, te = 1000000000, 1001000000
    result = itac.output_events(clustered_coinc, ts, te)

    assert isinstance(result, EventBuffer)


def test_output_events_empty_triggers():
    """Test output_events prints when no triggers."""
    itac = Itacacac(
        name="test_itacacac",
        sink_pad_names=["H1"],
        sample_rate=1024,
        trigger_finding_duration=1.0,
        snr_min=4.0,
        autocorrelation_banks={"H1": torch.zeros(1, 1, 10, dtype=torch.complex64)},
        autocorrelation_length_mask=None,
        autocorrelation_lengths={0: 10},
        template_ids=torch.tensor([[0]]),
        bankids_map={0: [0]},
        end_time_delta=torch.tensor([0.0]),
        template_durations=torch.tensor([[1.0]]),
        device="cpu",
        coincidence_threshold=0.0,
        min_instruments_candidates=1,
        strike_pad=None,
        stillsuit_pad="stillsuit",
    )

    # Empty clustered_coinc
    clustered_coinc = {
        "sngls": {"H1": {"time": torch.tensor([]), "shifted_time": torch.tensor([])}},
        "clustered_template_ids": torch.tensor([]),
        "clustered_template_durations": torch.tensor([]),
        "clustered_ifo_combs": torch.zeros(0, dtype=torch.int),
        "clustered_snr": torch.tensor([]),
        "clustered_bankids": [],
        "snr_ts_snippet_clustered": {"H1": None},
        "snr_ts_clustered": {"H1": None},
    }

    ts, te = 1000000000, 1001000000
    result = itac.output_events(clustered_coinc, ts, te)

    # Should handle empty case
    assert isinstance(result, EventBuffer)


def test_find_peaks_and_calculate_chisqs_with_mask():
    """Test find_peaks_and_calculate_chisqs with autocorrelation_length_mask."""
    mask = {"H1": torch.ones(1, 1, 20)}
    itac = Itacacac(
        name="test_itacacac",
        sink_pad_names=["H1"],
        sample_rate=1024,
        trigger_finding_duration=1.0,
        snr_min=4.0,
        autocorrelation_banks={"H1": torch.zeros(1, 1, 20, dtype=torch.complex64)},
        autocorrelation_length_mask=mask,
        autocorrelation_lengths={0: 20},
        template_ids=torch.tensor([[0]]),
        bankids_map={0: [0]},
        end_time_delta=torch.tensor([0.0]),
        template_durations=torch.tensor([[1.0]]),
        device="cpu",
        coincidence_threshold=0.0,
        min_instruments_candidates=1,
        strike_pad=None,
        stillsuit_pad="stillsuit",
    )

    # Create snr time series with shape (nsubbank, ntempmax*2, nsamples)
    # ntempmax*2 because real and imag are interleaved
    nsamples = 100
    snr_ts = {
        "H1": torch.randn(1, 2, nsamples),  # 1 subbank, 1 template * 2 (real/imag)
    }

    triggers = itac.find_peaks_and_calculate_chisqs(snr_ts)

    assert "peak_locations" in triggers
    assert "snrs" in triggers
    assert "chisqs" in triggers
    assert "H1" in triggers["snrs"]



"""Keep track of zero lag counts clustered over 10s fixed windows"""

# Copyright (C) 2016  Kipp Cannon, Chad Hanna
# Copyright (C) 2019  Patrick Godwin
# Copyright (C) 2025  Yun-Jing Huang


import argparse
import io
import json
import logging
import os
import threading
import time
import timeit
from collections import deque
from queue import Queue

from igwn_ligolw import utils as ligolw_utils
from lal.utils import CacheEntry
from ligo.scald import aggregator
from ligo.scald.io import kafka
from sgn.control import HTTPControl
from sgnligo.base import now as time_now
from strike.stats import far

from sgnl import events
from sgnl.dags.util import DEFAULT_BACKUP_DIR, T050017_filename

#
# =============================================================================
#
#                                 Command Line
#
# =============================================================================
#

program_name = "sgnl-ll-trigger-counter"


def xml_string(rstat):
    f = io.BytesIO()
    rstat.save_fileobj(f)
    f.seek(0)
    return f.read().decode("utf-8")


class ZeroLagCounts(events.EventProcessor):
    """
    A Class to keep track of zero lag counts clustered over 10s fixed windows
    """

    _name = "trigger_counter"

    def __init__(
        self,
        kafka_server,
        logger,
        sgn_control,
        output="zerolag_rankingstatpdf.xml.gz",
        output_period=3600,
        process_params=None,
        tag="test",
        topic="coinc",
        verbose=False,
    ):
        """
        fname is the name of the zerolag file read from disk to start the counting
        """
        if process_params is None:
            self.process_params = {}
        else:
            self.process_params = process_params
        self.logger = logger
        self.lock = threading.Lock()
        self.fname = output
        self.verbose = verbose
        self.time_since_last = time.time()
        self.output_period = output_period

        if os.access(ligolw_utils.local_path_from_url(self.fname), os.R_OK):
            self.zerolag_rankingstatpdf = far.RankingStatPDF.load(
                self.fname, verbose=verbose
            )
        else:
            raise ValueError("zerolag ranking stat PDF does not exist")
        self.counts_dict = {}

        # set up kafka client to continuously listen for messages
        events.EventProcessor.__init__(
            self,
            input_topic=f"sgnl.{tag}.{topic}",
            kafka_server=kafka_server,
            num_messages=10000,
            tag=f"{tag}-counter",
            send_heartbeats=True,
            heartbeat_cadence=60.0,
            heartbeat_topic=f"sgnl.{tag}.trigger_counter_heartbeat",
            sgn_control=sgn_control,
        )

        # store incoming messages in a deque before handling them
        self.msgs = deque()
        self.state_dict = {
            "zerolagxml": {"all": xml_string(self.zerolag_rankingstatpdf)}
        }
        HTTPControl.exchange_state(self._name, self.state_dict)

    def ingest(self, message):
        # keep list of incoming triggers
        if message.key():
            key = message.key().decode("utf-8").split(".")
        else:
            key = []
        self.msgs.append(
            kafka.Message(
                message.topic(),
                message.partition(),
                message.offset(),
                tuple(key),
                json.loads(message.value().decode("utf-8")),
                message.timestamp()[0],
            )
        )

    def handle(self):
        triggers = aggregator.parse_triggers(self.msgs)

        # clear out msgs deque to avoid double counting triggers
        self.msgs.clear()

        # store and reduce data for each job
        if triggers:
            start = timeit.default_timer()
            self.logger.debug("adding triggers")

            num = self.add_coincs(triggers)

            elapsed = timeit.default_timer() - start
            self.logger.debug("time to add %d triggers: %.1f s", num, elapsed)
        else:
            self.logger.debug("no triggers to process")

        if (time.time() - self.time_since_last) > self.output_period:
            self.logger.info("making snapshot")
            self.snapshot_output_url()
            self.time_since_last = time.time()

    def add_coincs(self, coincs):
        """
        Iterate over a list of coincs to extract their gps time and
        likelihood ratio value.  GPS times are rounded to the nearest 10s in order to
        cluster events. The max likelihood is tracked in the dictionary over the 10s
        window. Nothing is added to the zerolag histograms. The counts are purely
        internal. Only when a new zerolag histogram is requested will the zerolag
        histogram be updated. That makes it easier to manage late buffers.  No
        additional logic required.  The number of coincs added is returned for
        logging purposes
        """
        cnt = 0
        for c in coincs:
            # avoid trigs before LR assigned
            if "likelihood" in c:
                # equiv to 10s clustering
                key = round(c["end"], -1)
                lr = c["likelihood"]
                self.counts_dict[key] = max(self.counts_dict.setdefault(key, lr), lr)
                cnt += 1
        return cnt

    def __add_counts_to_zerolag_pdf(self, zlpdf):
        """
        Add the internal counts to a zerolag histogram. NOTE!!!! This
        should never be called on the internal zerolag histogram, only on a copy. That
        is why this method is not invoked unless someone is requesting an updated file
        """
        for lr in self.counts_dict.values():
            zlpdf.zero_lag_lr_lnpdf.count[lr,] += 1

    def snapshot_output_url(self):
        """
        Write a new XML doc with the internal counts added to disk with fname
        """
        with self.lock:
            HTTPControl.exchange_state(self._name, self.state_dict)
            zlpdf = self.zerolag_rankingstatpdf.copy()
            self.__add_counts_to_zerolag_pdf(zlpdf)
            zlpdf.save(
                self.fname,
                process_name=program_name,
                process_params=self.process_params,
                verbose=self.verbose,
            )
            self.state_dict["zerolagxml"]["all"] = xml_string(zlpdf)

        # save the same file to the backup dir as a precaution
        with self.lock:
            now = int(time_now())
            f = CacheEntry.from_T050017(self.fname)
            backup_dir = os.path.join(DEFAULT_BACKUP_DIR, os.path.dirname(self.fname))
            if not os.path.exists(backup_dir):
                os.makedirs(backup_dir)
            backup_fname = T050017_filename(
                f.observatory, f.description, (now, now), "xml.gz"
            )
            backup_fname = os.path.join(backup_dir, backup_fname)
            zlpdf.save(
                backup_fname,
                process_name=program_name,
                process_params=self.process_params,
                verbose=self.verbose,
            )


# Read command line options
def parse_command_line():

    parser = argparse.ArgumentParser(description="Online trigger counter")

    # directory to put everything in
    parser.add_argument(
        "--topic",
        default="coinc",
        help="Specify the topic where triggers are located.",
    )
    parser.add_argument(
        "--output-period",
        type=float,
        default=3600.0,
        help="Wait this many seconds between writing the output file (default = 3600)",
    )
    parser.add_argument(
        "--kafka-server",
        metavar="string",
        help="Sets the server url that the kafka topic is hosted on. Required.",
    )
    parser.add_argument(
        "--output",
        default="zerolag_rankingstatpdf.xml.gz",
        help="Choose the output file. Default zerolag_rankingstatpdf.xml.gz",
    )
    parser.add_argument(
        "--tag",
        metavar="string",
        default="test",
        help="Sets the name of the tag used. Default = 'test'",
    )
    parser.add_argument(
        "-v", "--verbose", default=False, action="store_true", help="Be verbose."
    )

    args = parser.parse_args()

    return args


#
# =============================================================================
#
#                                     Main
#
# =============================================================================
#


def main():
    options = parse_command_line()
    process_params = options.__dict__.copy()

    # set up logging
    log_level = logging.DEBUG if options.verbose else logging.INFO
    logging.basicConfig(
        format="%(asctime)s | sgnl-ll-trigger-counter : %(levelname)s : %(message)s"
    )
    logger = logging.getLogger("trigger_counter")
    logger.setLevel(log_level)

    port = 40000  # FIXME: do we use other ports??
    HTTPControl.port = port
    HTTPControl.tag = options.tag
    HTTPControl.get_queues["trigger_counter"] = Queue(1)
    HTTPControl.post_queues["trigger_counter"] = Queue(1)
    registry_file = "sgnl_trigger_counter_registry.txt"
    with HTTPControl(registry_file=registry_file) as sgn_control:
        ZLC = ZeroLagCounts(
            kafka_server=options.kafka_server,
            logger=logger,
            sgn_control=sgn_control,
            output=options.output,
            output_period=options.output_period,
            process_params=process_params,
            tag=options.tag,
            topic=options.topic,
            verbose=options.verbose,
        )

        # start an infinite loop to keep updating and aggregating data
        logger.info("starting up...")
        ZLC.start()



"""Tests for sgnl.bin.segment_ops CLI"""

import sys
from unittest import mock

import igwn_segments as segments
import pytest

from sgnl.bin import segment_ops


class TestParseCommandLine:
    """Tests for parse_command_line function."""

    def test_single_input_file(self):
        """Test parsing with single input file."""
        with mock.patch.object(sys, "argv", ["segment_ops", "input.xml"]):
            args = segment_ops.parse_command_line()

        assert args.inputs == ["input.xml"]
        assert args.output == "segments_out.xml.gz"
        assert args.segment_name == "datasegments"

    def test_multiple_input_files(self):
        """Test parsing with multiple input files."""
        with mock.patch.object(
            sys, "argv", ["segment_ops", "input1.xml", "input2.xml", "input3.xml"]
        ):
            args = segment_ops.parse_command_line()

        assert args.inputs == ["input1.xml", "input2.xml", "input3.xml"]

    def test_custom_output_file(self):
        """Test parsing with custom output file."""
        with mock.patch.object(
            sys, "argv", ["segment_ops", "input.xml", "-o", "custom_output.xml.gz"]
        ):
            args = segment_ops.parse_command_line()

        assert args.output == "custom_output.xml.gz"

    def test_custom_segment_name(self):
        """Test parsing with custom segment name."""
        with mock.patch.object(
            sys, "argv", ["segment_ops", "input.xml", "-n", "custom_segments"]
        ):
            args = segment_ops.parse_command_line()

        assert args.segment_name == "custom_segments"

    def test_union_operation(self):
        """Test parsing with union operation."""
        with mock.patch.object(
            sys,
            "argv",
            ["segment_ops", "input1.xml", "input2.xml", "--operation", "union"],
        ):
            args = segment_ops.parse_command_line()

        assert args.operation == "union"

    def test_intersection_operation(self):
        """Test parsing with intersection operation."""
        with mock.patch.object(
            sys,
            "argv",
            ["segment_ops", "input1.xml", "input2.xml", "--operation", "intersection"],
        ):
            args = segment_ops.parse_command_line()

        assert args.operation == "intersection"

    def test_diff_operation(self):
        """Test parsing with diff operation."""
        with mock.patch.object(
            sys,
            "argv",
            ["segment_ops", "input1.xml", "input2.xml", "--operation", "diff"],
        ):
            args = segment_ops.parse_command_line()

        assert args.operation == "diff"

    def test_gps_bounds(self):
        """Test parsing with GPS bounds."""
        with mock.patch.object(
            sys,
            "argv",
            [
                "segment_ops",
                "input.xml",
                "--gps-start",
                "1000000000",
                "--gps-end",
                "1000010000",
            ],
        ):
            args = segment_ops.parse_command_line()

        assert args.gps_start == 1000000000.0
        assert args.gps_end == 1000010000.0

    def test_min_length(self):
        """Test parsing with min-length."""
        with mock.patch.object(
            sys, "argv", ["segment_ops", "input.xml", "--min-length", "512"]
        ):
            args = segment_ops.parse_command_line()

        assert args.min_length == 512.0

    def test_trim(self):
        """Test parsing with trim."""
        with mock.patch.object(
            sys, "argv", ["segment_ops", "input.xml", "--trim", "100"]
        ):
            args = segment_ops.parse_command_line()

        assert args.trim == 100.0

    def test_verbose_flag(self):
        """Test parsing with verbose flag."""
        with mock.patch.object(sys, "argv", ["segment_ops", "input.xml", "-v"]):
            args = segment_ops.parse_command_line()

        assert args.verbose is True

    def test_diff_requires_two_files(self):
        """Test that diff operation requires exactly two files."""
        with mock.patch.object(
            sys, "argv", ["segment_ops", "input.xml", "--operation", "diff"]
        ):
            with pytest.raises(SystemExit):
                segment_ops.parse_command_line()

    def test_diff_with_three_files_fails(self):
        """Test that diff operation with three files fails."""
        with mock.patch.object(
            sys,
            "argv",
            [
                "segment_ops",
                "input1.xml",
                "input2.xml",
                "input3.xml",
                "--operation",
                "diff",
            ],
        ):
            with pytest.raises(SystemExit):
                segment_ops.parse_command_line()

    def test_union_requires_two_files(self):
        """Test that union operation requires at least two files."""
        with mock.patch.object(
            sys, "argv", ["segment_ops", "input.xml", "--operation", "union"]
        ):
            with pytest.raises(SystemExit):
                segment_ops.parse_command_line()

    def test_intersection_requires_two_files(self):
        """Test that intersection operation requires at least two files."""
        with mock.patch.object(
            sys, "argv", ["segment_ops", "input.xml", "--operation", "intersection"]
        ):
            with pytest.raises(SystemExit):
                segment_ops.parse_command_line()


class TestMain:
    """Tests for main function."""

    def test_main_single_file(self, tmp_path):
        """Test main with single input file."""
        # Create input segment file
        input_segs = segments.segmentlistdict()
        input_segs["H1"] = segments.segmentlist([segments.segment(0, 1000)])

        input_file = tmp_path / "input.xml.gz"
        output_file = tmp_path / "output.xml.gz"

        from sgnl.dags.segments import write_segments

        write_segments(input_segs, output=str(input_file))

        with mock.patch.object(
            sys, "argv", ["segment_ops", str(input_file), "-o", str(output_file)]
        ):
            segment_ops.main()

        assert output_file.exists()

    def test_main_with_gps_bounds(self, tmp_path):
        """Test main with GPS bounding."""
        input_segs = segments.segmentlistdict()
        input_segs["H1"] = segments.segmentlist([segments.segment(0, 1000)])

        input_file = tmp_path / "input.xml.gz"
        output_file = tmp_path / "output.xml.gz"

        from sgnl.dags.segments import load_segment_file, write_segments

        write_segments(input_segs, output=str(input_file))

        with mock.patch.object(
            sys,
            "argv",
            [
                "segment_ops",
                str(input_file),
                "-o",
                str(output_file),
                "--gps-start",
                "100",
                "--gps-end",
                "500",
            ],
        ):
            segment_ops.main()

        result = load_segment_file(str(output_file))
        assert result["H1"][0] == segments.segment(100, 500)

    def test_main_with_min_length(self, tmp_path):
        """Test main with minimum length filtering."""
        input_segs = segments.segmentlistdict()
        input_segs["H1"] = segments.segmentlist(
            [
                segments.segment(0, 100),  # too short
                segments.segment(500, 1500),  # long enough
            ]
        )

        input_file = tmp_path / "input.xml.gz"
        output_file = tmp_path / "output.xml.gz"

        from sgnl.dags.segments import load_segment_file, write_segments

        write_segments(input_segs, output=str(input_file))

        with mock.patch.object(
            sys,
            "argv",
            [
                "segment_ops",
                str(input_file),
                "-o",
                str(output_file),
                "--min-length",
                "512",
            ],
        ):
            segment_ops.main()

        result = load_segment_file(str(output_file))
        assert len(result["H1"]) == 1
        assert result["H1"][0] == segments.segment(500, 1500)

    def test_main_with_trim(self, tmp_path):
        """Test main with segment contraction."""
        input_segs = segments.segmentlistdict()
        input_segs["H1"] = segments.segmentlist([segments.segment(0, 1000)])

        input_file = tmp_path / "input.xml.gz"
        output_file = tmp_path / "output.xml.gz"

        from sgnl.dags.segments import load_segment_file, write_segments

        write_segments(input_segs, output=str(input_file))

        with mock.patch.object(
            sys,
            "argv",
            [
                "segment_ops",
                str(input_file),
                "-o",
                str(output_file),
                "--trim",
                "50",
            ],
        ):
            segment_ops.main()

        result = load_segment_file(str(output_file))
        assert result["H1"][0] == segments.segment(50, 950)

    def test_main_union_operation(self, tmp_path):
        """Test main with union operation."""
        segs1 = segments.segmentlistdict()
        segs1["H1"] = segments.segmentlist([segments.segment(0, 500)])

        segs2 = segments.segmentlistdict()
        segs2["H1"] = segments.segmentlist([segments.segment(400, 1000)])

        input_file1 = tmp_path / "input1.xml.gz"
        input_file2 = tmp_path / "input2.xml.gz"
        output_file = tmp_path / "output.xml.gz"

        from sgnl.dags.segments import load_segment_file, write_segments

        write_segments(segs1, output=str(input_file1))
        write_segments(segs2, output=str(input_file2))

        with mock.patch.object(
            sys,
            "argv",
            [
                "segment_ops",
                str(input_file1),
                str(input_file2),
                "-o",
                str(output_file),
                "--operation",
                "union",
            ],
        ):
            segment_ops.main()

        result = load_segment_file(str(output_file))
        assert len(result["H1"]) == 1
        assert result["H1"][0] == segments.segment(0, 1000)

    def test_main_intersection_operation(self, tmp_path):
        """Test main with intersection operation."""
        segs1 = segments.segmentlistdict()
        segs1["H1"] = segments.segmentlist([segments.segment(0, 700)])

        segs2 = segments.segmentlistdict()
        segs2["H1"] = segments.segmentlist([segments.segment(300, 1000)])

        input_file1 = tmp_path / "input1.xml.gz"
        input_file2 = tmp_path / "input2.xml.gz"
        output_file = tmp_path / "output.xml.gz"

        from sgnl.dags.segments import load_segment_file, write_segments

        write_segments(segs1, output=str(input_file1))
        write_segments(segs2, output=str(input_file2))

        with mock.patch.object(
            sys,
            "argv",
            [
                "segment_ops",
                str(input_file1),
                str(input_file2),
                "-o",
                str(output_file),
                "--operation",
                "intersection",
            ],
        ):
            segment_ops.main()

        result = load_segment_file(str(output_file))
        assert len(result["H1"]) == 1
        assert result["H1"][0] == segments.segment(300, 700)

    def test_main_diff_operation(self, tmp_path):
        """Test main with diff operation."""
        segs1 = segments.segmentlistdict()
        segs1["H1"] = segments.segmentlist([segments.segment(0, 1000)])

        segs2 = segments.segmentlistdict()
        segs2["H1"] = segments.segmentlist([segments.segment(300, 700)])

        input_file1 = tmp_path / "input1.xml.gz"
        input_file2 = tmp_path / "input2.xml.gz"
        output_file = tmp_path / "output.xml.gz"

        from sgnl.dags.segments import load_segment_file, write_segments

        write_segments(segs1, output=str(input_file1))
        write_segments(segs2, output=str(input_file2))

        with mock.patch.object(
            sys,
            "argv",
            [
                "segment_ops",
                str(input_file1),
                str(input_file2),
                "-o",
                str(output_file),
                "--operation",
                "diff",
            ],
        ):
            segment_ops.main()

        result = load_segment_file(str(output_file))
        assert len(result["H1"]) == 2
        assert result["H1"][0] == segments.segment(0, 300)
        assert result["H1"][1] == segments.segment(700, 1000)

    def test_main_verbose(self, tmp_path, capsys):
        """Test main with verbose output."""
        input_segs = segments.segmentlistdict()
        input_segs["H1"] = segments.segmentlist([segments.segment(0, 1000)])

        input_file = tmp_path / "input.xml.gz"
        output_file = tmp_path / "output.xml.gz"

        from sgnl.dags.segments import write_segments

        write_segments(input_segs, output=str(input_file))

        with mock.patch.object(
            sys,
            "argv",
            ["segment_ops", str(input_file), "-o", str(output_file), "-v"],
        ):
            segment_ops.main()

        captured = capsys.readouterr()
        assert "Loading segments" in captured.out
        assert "Saving output" in captured.out

    def test_main_combined_operations(self, tmp_path):
        """Test main with multiple operations combined."""
        input_segs = segments.segmentlistdict()
        input_segs["H1"] = segments.segmentlist(
            [
                segments.segment(0, 100),  # will be filtered
                segments.segment(1000, 3000),  # will be bounded and trimmed
            ]
        )

        input_file = tmp_path / "input.xml.gz"
        output_file = tmp_path / "output.xml.gz"

        from sgnl.dags.segments import load_segment_file, write_segments

        write_segments(input_segs, output=str(input_file))

        with mock.patch.object(
            sys,
            "argv",
            [
                "segment_ops",
                str(input_file),
                "-o",
                str(output_file),
                "--gps-start",
                "500",
                "--gps-end",
                "2500",
                "--min-length",
                "512",
                "--trim",
                "100",
            ],
        ):
            segment_ops.main()

        result = load_segment_file(str(output_file))
        # After bounding: [1000, 2500]
        # After filtering (>= 512): [1000, 2500]
        # After trimming by 100: [1100, 2400]
        assert len(result["H1"]) == 1
        assert result["H1"][0] == segments.segment(1100, 2400)

    def test_main_verbose_with_all_operations(self, tmp_path, capsys):
        """Test verbose output for all operations."""
        segs1 = segments.segmentlistdict()
        segs1["H1"] = segments.segmentlist([segments.segment(0, 5000)])

        segs2 = segments.segmentlistdict()
        segs2["H1"] = segments.segmentlist([segments.segment(1000, 4000)])

        input_file1 = tmp_path / "input1.xml.gz"
        input_file2 = tmp_path / "input2.xml.gz"
        output_file = tmp_path / "output.xml.gz"

        from sgnl.dags.segments import write_segments

        write_segments(segs1, output=str(input_file1))
        write_segments(segs2, output=str(input_file2))

        with mock.patch.object(
            sys,
            "argv",
            [
                "segment_ops",
                str(input_file1),
                str(input_file2),
                "-o",
                str(output_file),
                "-v",
                "--operation",
                "union",
                "--gps-start",
                "500",
                "--gps-end",
                "4500",
                "--min-length",
                "100",
                "--trim",
                "50",
            ],
        ):
            segment_ops.main()

        captured = capsys.readouterr()
        assert "Applying set operation: union" in captured.out
        assert "Applying GPS bounding" in captured.out
        assert "Filtering segments shorter than" in captured.out
        assert "Contracting segments by" in captured.out



"""A program to plot a psd such as one generated by sgnl-reference-psd"""

# Copyright (C) 2012 Chad Hanna
# Copyright (C) 2024 Yun-Jing Huang


import os
from argparse import ArgumentParser

from sgnl.plots.util import set_matplotlib_cache_directory

set_matplotlib_cache_directory()
import matplotlib  # noqa: E402

from sgnl.plots.psd import plot_psds  # noqa: E402
from sgnl.psd import read_psd  # noqa: E402

matplotlib.rcParams.update(
    {
        "font.size": 10.0,
        "axes.titlesize": 10.0,
        "axes.labelsize": 10.0,
        "xtick.labelsize": 8.0,
        "ytick.labelsize": 8.0,
        "legend.fontsize": 8.0,
        "figure.dpi": 300,
        "savefig.dpi": 300,
        "text.usetex": False,
        "path.simplify": True,
    }
)


def parse_command_line():
    parser = ArgumentParser()
    parser.add_argument("filenames", nargs="+")
    parser.add_argument(
        "-o",
        "--output",
        help="Set plot filename (default = replace input file's extension with .png).",
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose.")
    options = parser.parse_args()
    filenames = options.filenames

    if not filenames:
        raise ValueError("must supply at least one input filename")
    if options.output and len(filenames) > 1:
        raise ValueError("must supply only one input file when setting --output")

    return options


def main():
    options = parse_command_line()

    for fname in options.filenames:
        if options.output:
            outname = options.output
        else:
            outname = os.path.join(os.path.splitext(fname)[0], ".png")

        fig = plot_psds(read_psd(fname, verbose=options.verbose), plot_width=2400)
        fig.savefig(outname)



# Copyright (C) 2024  Cort Posnansky (cort.posnansky@ligo.org)
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


import argparse
import os
import pathlib
import sys

from ezdag import DAG

from sgnl.bin import inspiral_bank_splitter, inspiral_set_svdbin_option
from sgnl.dags import layers
from sgnl.dags.config import build_config, create_time_bins
from sgnl.dags.util import (
    DataCache,
    DataType,
    add_osdf_support_to_layer,
    load_svd_options,
    mchirp_range_to_bins,
    osdf_flatten_frame_cache,
)


def parse_command_line(args: list[str] | None = None) -> argparse.Namespace:
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(
        description="Generate a sgnl dag from a config file"
    )
    parser.add_argument(
        "-i",
        "--init",
        action="store_true",
        help="Run initialization steps needed before making a dag. Cannot be "
        "provided at same time as --workflow.",
    )
    parser.add_argument(
        "-c", "--config", type=str, required=True, help="The config file to load"
    )
    parser.add_argument(
        "-w", "--workflow", type=str, help="The type of dag to generate"
    )
    parser.add_argument(
        "--dag-dir",
        type=str,
        default=".",
        help="The directory in which to write the dag",
    )
    parser.add_argument("--dag-name", type=str, default=None, help="A name for the dag")
    parser.add_argument(
        "--force-segments",
        action="store_true",
        help="Force overwrite of segments file.",
    )
    parsed_args = parser.parse_args(args)

    if parsed_args.dag_name:
        if "/" in parsed_args.dag_name:
            raise ValueError(
                "The dag name must not be a path. Use --dag-dir to put the dag in a"
                " different directory."
            )
        if parsed_args.dag_name.endswith(".dag"):
            raise ValueError(
                'The given dag name ends with ".dag". This is unnecessary because it'
                " will be appended to the file name automatically."
            )

    if parsed_args.init is True and parsed_args.workflow is not None:
        raise ValueError("Cannot specify both --init and --workflow.")
    elif parsed_args.init is False and parsed_args.workflow is None:
        raise ValueError("Must specify either --init or --workflow.")

    return parsed_args


# FIXME We're planning to handle OSDF support better later, this function can
# be replaced or removed then
def prepare_osdf_support(config, dag_dir):
    """
    Prepare config for OSDF frame support by flattening any OSDF frame caches.
    Call this once after loading config, before any workflow generation.
    """
    if config.source and config.source.frame_cache:
        flat_cache_path = os.path.join(dag_dir, "flat_frame_cache.cache")
        result = osdf_flatten_frame_cache(
            config.source.frame_cache, new_cache_name=flat_cache_path
        )
        (
            config.source.frames_in_osdf,
            config.source.transfer_frame_cache,
        ) = result

    if config.source and config.source.inj_frame_cache:
        flat_inj_cache_path = os.path.join(dag_dir, "flat_inj_frame_cache.cache")
        result = osdf_flatten_frame_cache(
            config.source.inj_frame_cache, new_cache_name=flat_inj_cache_path
        )
        (
            config.source.inj_frames_in_osdf,
            config.source.transfer_inj_frame_cache,
        ) = result


def main():
    args = parse_command_line()
    config = build_config(args.config, args.dag_dir, force_segments=args.force_segments)
    if args.init:
        # NOTE bank_name, output_full_bank_file, psd, psd_xml, and
        # psdinterp bank-splitter options are currently not implemented
        # options here. User will need to manually run bank splitter for
        # this functionality at this time
        arguments = {
            "bank_name": None,
            "f_final": config.svd.max_f_final,
            "f_low": config.svd.f_low,
            "filename": config.paths.template_bank,
            "instrument": config.instruments,
            "n": config.svd.num_split_templates,
            "num_banks": (
                config.svd.num_banks
                if isinstance(config.svd.num_banks, list)
                else [config.svd.num_banks]
            ),
            "output_path": os.path.join(config.paths.input_data, "split_bank"),
            "stats_file": config.svd.option_file,
            "verbose": True,
        }

        if config.svd.sort_by:
            arguments["sort_by"] = config.svd.sort_by

        if config.svd.sort_by == "chi":
            arguments["group_by_chi"] = config.svd.num_chi_bins

        if config.svd.sort_by == "mu":
            arguments["group_by_mu"] = config.svd.num_mu_bins

        if config.svd.overlap:
            arguments["overlap"] = config.svd.overlap

        inspiral_bank_splitter.split_bank(
            **arguments,
            approximants=inspiral_bank_splitter.split_approximant_strings(
                config.svd.approximant
            ),
            argument_dict=arguments
            | {
                "approximants": config.svd.approximant
            },  # argument dict needs the approx strings
        )

        inspiral_set_svdbin_option.set_svdbin_option(config)
        return

    svd_bins, svd_stats = load_svd_options(config.svd.option_file, config.svd)
    max_duration = max(svd_bin["max_dur"] for svd_bin in svd_stats.bins.values())
    filter_start_pad = 16 * config.psd.fft_length + max_duration
    create_time_bins(config, start_pad=filter_start_pad)

    # Prepare OSDF support once for all workflows
    # FIXME We're planning to handle OSDF support better later, this function
    # can be replaced or removed then
    if not args.workflow == "rank":
        prepare_osdf_support(config, args.dag_dir)

    if args.dag_name:
        dag_name = args.dag_name
    else:
        dag_name = f"sgnl_{args.workflow}"

    # Start building the dag
    dag = DAG(dag_name)
    if args.workflow == "test":
        # FIXME Delete this workflow
        dag.attach(layers.test(config.echo, config.condor))

    elif args.workflow == "psd":
        # Reference PSD layer
        ref_psd_cache = DataCache.generate(
            DataType.REFERENCE_PSD,
            config.ifo_combos,
            config.time_bins,
            root=config.paths.input_data,
        )

        ref_psd_layer = layers.reference_psd(
            filter_config=config.filter,
            psd_config=config.psd,
            source_config=config.source,
            condor_config=config.condor,
            ref_psd_cache=ref_psd_cache,
        )

        # Add OSDF support if needed
        # FIXME We're planning to handle OSDF support better later, this
        # function can be replaced or removed then
        add_osdf_support_to_layer(ref_psd_layer, config.source)

        dag.attach(ref_psd_layer)

        # Median PSD layer
        median_psd_cache = DataCache.generate(
            DataType.MEDIAN_PSD,
            config.all_ifos,
            config.span,
            root=config.paths.input_data,
        )

        dag.attach(
            layers.median_psd(
                psd_config=config.psd,
                condor_config=config.condor,
                ref_psd_cache=ref_psd_cache,
                median_psd_cache=median_psd_cache,
            )
        )

    elif args.workflow == "svd":
        split_bank_cache = DataCache.find(
            DataType.SPLIT_BANK, svd_bins="*", subtype="*", root=config.paths.input_data
        )
        median_psd_cache = DataCache.find(
            DataType.MEDIAN_PSD, root=config.paths.input_data
        )

        svd_cache = DataCache.generate(
            DataType.SVD_BANK,
            config.ifos,
            config.span,
            svd_bins=svd_bins,
            root=config.paths.input_data,
        )

        dag.attach(
            layers.svd_bank(
                svd_config=config.svd,
                condor_config=config.condor,
                all_ifos=list(sorted(config.all_ifos)),
                split_bank_cache=split_bank_cache,
                median_psd_cache=median_psd_cache,
                svd_cache=svd_cache,
                svd_bins=svd_bins,
                svd_stats=svd_stats,
            )
        )

    elif args.workflow == "filter":
        if not config.paths.filter_dir:
            config.paths.filter_dir = config.paths.storage

        ref_psd_cache = DataCache.find(
            DataType.REFERENCE_PSD, root=config.paths.input_data
        )
        svd_bank_cache = DataCache.find(
            DataType.SVD_BANK, root=config.paths.input_data, svd_bins="*"
        )

        lr_cache = DataCache.generate(
            DataType.LIKELIHOOD_RATIO,
            config.ifo_combos,
            config.time_bins,
            svd_bins=svd_bins,
            root=config.paths.filter_dir,
        )

        trigger_cache = DataCache.generate(
            DataType.TRIGGERS,
            config.ifo_combos,
            config.time_bins,
            svd_bins=svd_bins,
            root=config.paths.filter_dir,
        )

        filter_layer = layers.filter(
            psd_config=config.psd,
            svd_config=config.svd,
            filter_config=config.filter,
            source_config=config.source,
            condor_config=config.condor,
            ref_psd_cache=ref_psd_cache,
            svd_bank_cache=svd_bank_cache,
            lr_cache=lr_cache,
            trigger_cache=trigger_cache,
            svd_stats=svd_stats,
            min_instruments=config.filter.min_instruments_candidates,
        )

        # Add OSDF support if needed
        # FIXME We're planning to handle OSDF support better later, this
        # function can be replaced or removed then
        add_osdf_support_to_layer(filter_layer, config.source)

        dag.attach(filter_layer)

        marg_lr_cache = DataCache.generate(
            DataType.MARG_LIKELIHOOD_RATIO,
            config.all_ifos,
            config.span,
            svd_bins=lr_cache.groupby("bin").keys(),
            root=config.paths.filter_dir,
        )
        layer = layers.marginalize_likelihood_ratio(
            condor_config=config.condor,
            lr_cache=lr_cache,
            marg_lr_cache=marg_lr_cache,
        )

        dag.attach(layer)

    elif args.workflow == "injection-filter":
        if not config.paths.injection_dir:
            config.paths.injection_dir = config.paths.storage

        ref_psd_cache = DataCache.find(
            DataType.REFERENCE_PSD, root=config.paths.input_data
        )
        svd_bank_cache = DataCache.find(
            DataType.SVD_BANK, root=config.paths.input_data, svd_bins="*"
        )

        trigger_cache = DataCache(DataType.TRIGGERS)

        for inj_name, inj_args in config.injections.filter.items():
            min_mchirp, max_mchirp = map(float, inj_args["range"].split(":"))
            svd_bins = mchirp_range_to_bins(min_mchirp, max_mchirp, svd_stats)
            trigger_cache += DataCache.generate(
                DataType.TRIGGERS,
                config.ifo_combos,
                config.time_bins,
                svd_bins=svd_bins,
                subtype=inj_name,
                root=config.paths.injection_dir,
            )

        injection_filter_layer = layers.injection_filter(
            psd_config=config.psd,
            svd_config=config.svd,
            filter_config=config.filter,
            injection_config=config.injections,
            source_config=config.source,
            condor_config=config.condor,
            ref_psd_cache=ref_psd_cache,
            svd_bank_cache=svd_bank_cache,
            trigger_cache=trigger_cache,
            svd_stats=svd_stats,
            min_instruments=config.filter.min_instruments_candidates,
        )

        # Add OSDF support if needed
        # FIXME We're planning to handle OSDF support better later, this
        # function can be replaced or removed then
        add_osdf_support_to_layer(
            injection_filter_layer, config.source, is_injection_workflow=True
        )

        dag.attach(injection_filter_layer)

    elif args.workflow == "rank":
        # FIXME: add online chunks

        # initialize empty caches, to which we will
        # add discovered data products
        marg_lr_cache = DataCache(DataType.MARG_LIKELIHOOD_RATIO)
        triggers_cache = DataCache(DataType.TRIGGERS)
        if config.injections:
            inj_triggers_cache = DataCache(DataType.TRIGGERS)
        prior_cache = DataCache(DataType.PRIOR_LIKELIHOOD_RATIO)

        svd_bank_cache = DataCache.find(
            DataType.SVD_BANK, root=config.paths.input_data, svd_bins="*"
        )
        marg_lr_cache += DataCache.find(
            DataType.MARG_LIKELIHOOD_RATIO, root=config.paths.filter_dir, svd_bins="*"
        )
        triggers_cache += DataCache.find(
            DataType.TRIGGERS, root=config.paths.filter_dir, svd_bins="*"
        )
        if config.injections:
            inj_triggers_cache += DataCache.find(
                DataType.TRIGGERS,
                root=config.paths.injection_dir,
                svd_bins="*",
                subtype="*",
            )

        prior_cache = DataCache.generate(
            DataType.PRIOR_LIKELIHOOD_RATIO,
            config.all_ifos,
            config.span,
            svd_bins=svd_bins,
            root=config.paths.rank_dir,
        )

        layer = layers.create_prior(
            filter_config=config.filter,
            condor_config=config.condor,
            prior_config=config.prior,
            coincidence_threshold=config.filter.coincidence_threshold,
            svd_bank_cache=svd_bank_cache,
            prior_cache=prior_cache,
            ifos=config.ifos,
            min_instruments=config.filter.min_instruments_candidates,
            svd_stats=svd_stats,
        )
        dag.attach(layer)

        marg_lr_prior_cache = DataCache.generate(
            DataType.MARG_LIKELIHOOD_RATIO_PRIOR,
            config.all_ifos,
            config.span,
            svd_bins=svd_bins,
            root=config.paths.rank_dir,
        )

        layer = layers.marginalize_likelihood_ratio(
            condor_config=config.condor,
            lr_cache=marg_lr_cache,
            marg_lr_cache=marg_lr_prior_cache,
            prior_cache=prior_cache,
        )
        dag.attach(layer)

        time_clustered_triggers_cache = DataCache(DataType.CLUSTERED_TRIGGERS)
        time_clustered_triggers_cache += DataCache.generate(
            DataType.CLUSTERED_TRIGGERS,
            config.all_ifos,
            config.span,
            svd_bins=svd_bins,
            root=config.paths.rank_dir,
        )
        if config.injections:
            for inj_name, inj_args in config.injections.filter.items():
                min_mchirp, max_mchirp = map(float, inj_args["range"].split(":"))
                svd_bins_inj = mchirp_range_to_bins(min_mchirp, max_mchirp, svd_stats)
                time_clustered_triggers_cache += DataCache.generate(
                    DataType.CLUSTERED_TRIGGERS,
                    config.all_ifos,
                    config.span,
                    svd_bins=svd_bins_inj,
                    subtype=inj_name,
                    root=config.paths.rank_dir,
                )

        all_clustered_triggers_cache = (
            triggers_cache + inj_triggers_cache if config.injections else triggers_cache
        )
        layer = layers.add_trigger_dbs(
            condor_config=config.condor,
            filter_config=config.filter,
            trigger_cache=all_clustered_triggers_cache,
            clustered_trigger_cache=time_clustered_triggers_cache,
            column="network_chisq_weighted_snr",
            window=0.1,
        )
        dag.attach(layer)

        lr_triggers_cache = DataCache(DataType.LR_TRIGGERS)
        lr_triggers_cache += DataCache.generate(
            DataType.LR_TRIGGERS,
            config.all_ifos,
            config.span,
            svd_bins=svd_bins,
            root=config.paths.rank_dir,
        )
        if config.injections:
            for inj_name, inj_args in config.injections.filter.items():
                min_mchirp, max_mchirp = map(float, inj_args["range"].split(":"))
                svd_bins_inj = mchirp_range_to_bins(min_mchirp, max_mchirp, svd_stats)
                lr_triggers_cache += DataCache.generate(
                    DataType.LR_TRIGGERS,
                    config.all_ifos,
                    config.span,
                    svd_bins=svd_bins_inj,
                    subtype=inj_name,
                    root=config.paths.rank_dir,
                )
        layer = layers.assign_likelihood(
            condor_config=config.condor,
            filter_config=config.filter,
            prior_config=config.prior,
            trigger_cache=time_clustered_triggers_cache,
            lr_cache=marg_lr_prior_cache,
            lr_trigger_cache=lr_triggers_cache,
            svd_stats=svd_stats,
        )
        dag.attach(layer)

        clustered_lr_triggers_cache = DataCache.generate(
            DataType.LR_TRIGGERS,
            config.all_ifos,
            config.span,
            svd_bins=f"{min(svd_bins)}_{max(svd_bins)}",
            root=config.paths.rank_dir,
        )
        if config.injections:
            for inj_name, inj_args in config.injections.filter.items():
                min_mchirp, max_mchirp = map(float, inj_args["range"].split(":"))
                svd_bins_inj = mchirp_range_to_bins(min_mchirp, max_mchirp, svd_stats)
                clustered_lr_triggers_cache += DataCache.generate(
                    DataType.LR_TRIGGERS,
                    config.all_ifos,
                    config.span,
                    svd_bins=f"{min(svd_bins_inj)}_{max(svd_bins_inj)}",
                    subtype=inj_name,
                    root=config.paths.rank_dir,
                )
        layer = layers.merge_and_reduce(
            condor_config=config.condor,
            filter_config=config.filter,
            trigger_cache=lr_triggers_cache,
            clustered_trigger_cache=clustered_lr_triggers_cache,
            column="likelihood",
            window=4,
        )
        dag.attach(layer)

        num_jobs = (
            config.rank.calc_pdf_jobs if config.rank.calc_pdf_jobs else 1
        )  # jobs per bin
        if num_jobs == 1:
            cache_svd_bins = svd_bins
        else:
            cache_svd_bins = []
            for svd_bin in svd_bins:
                for i in range(num_jobs):
                    cache_svd_bins.append(svd_bin + "_" + str(i))
        pdf_cache = DataCache.generate(
            DataType.RANK_STAT_PDFS,
            config.all_ifos,
            config.span,
            svd_bins=cache_svd_bins,
            root=config.paths.rank_dir,
        )
        layer = layers.calc_pdf(
            condor_config=config.condor,
            prior_config=config.prior,
            rank_config=config.rank,
            config_svd_bins=svd_bins,
            lr_cache=marg_lr_prior_cache,
            pdf_cache=pdf_cache,
            svd_stats=svd_stats,
        )
        dag.attach(layer)

        extinct_pdf_cache = DataCache.generate(
            DataType.RANK_STAT_PDFS,
            config.all_ifos,
            config.span,
            svd_bins=svd_bins,
            root=config.paths.rank_dir,
        )
        layer = layers.extinct_bin(
            condor_config=config.condor,
            event_config_file=config.filter.event_config_file,
            pdf_cache=pdf_cache,
            trigger_cache=lr_triggers_cache,
            extinct_cache=extinct_pdf_cache,
        )
        dag.attach(layer)

        marg_pdf_cache = DataCache.generate(
            DataType.RANK_STAT_PDFS,
            config.all_ifos,
            config.span,
            root=config.paths.rank_dir,
        )
        layer_list = layers.marginalize_pdf(
            condor_config=config.condor,
            rank_config=config.rank,
            rank_dir=config.paths.rank_dir,
            all_ifos=config.all_ifos,
            span=config.span,
            pdf_cache=extinct_pdf_cache,
            marg_pdf_cache=marg_pdf_cache,
        )
        for layer in layer_list:
            dag.attach(layer)

        post_pdf_cache = DataCache.generate(
            DataType.POST_RANK_STAT_PDFS,
            config.all_ifos,
            config.span,
            root=config.paths.rank_dir,
        )
        far_trigger_cache = DataCache.generate(
            DataType.FAR_TRIGGERS,
            config.all_ifos,
            config.span,
            svd_bins=f"{min(svd_bins)}_{max(svd_bins)}",
            root=config.paths.rank_dir,
        )
        if config.injections:
            for inj_name in config.injections.filter:
                far_trigger_cache += DataCache.generate(
                    DataType.FAR_TRIGGERS,
                    config.all_ifos,
                    config.span,
                    svd_bins=f"{min(svd_bins)}_{max(svd_bins)}",
                    subtype=inj_name,
                    root=config.paths.rank_dir,
                )

        layer_list = layers.assign_far(
            condor_config=config.condor,
            event_config_file=config.filter.event_config_file,
            trigger_cache=clustered_lr_triggers_cache,
            marg_pdf_cache=marg_pdf_cache,
            post_pdf_cache=post_pdf_cache,
            far_trigger_cache=far_trigger_cache,
        )
        for layer in layer_list:
            dag.attach(layer)

        seg_far_trigger_cache = DataCache.generate(
            DataType.SEGMENTS_FAR_TRIGGERS,
            config.all_ifos,
            config.span,
            svd_bins=f"{min(svd_bins)}_{max(svd_bins)}",
            root=config.paths.rank_dir,
        )
        if config.injections:
            for inj_name in config.injections.filter:
                seg_far_trigger_cache += DataCache.generate(
                    DataType.SEGMENTS_FAR_TRIGGERS,
                    config.all_ifos,
                    config.span,
                    svd_bins=f"{min(svd_bins)}_{max(svd_bins)}",
                    subtype=inj_name,
                    root=config.paths.rank_dir,
                )

        if not os.path.exists(config.summary.webdir):
            os.makedirs(config.summary.webdir)

        layer_list = layers.summary_page(
            condor_config=config.condor,
            event_config_file=config.filter.event_config_file,
            segments_file=config.source.frame_segments_file,
            segments_name=config.source.frame_segments_name,
            webdir=config.summary.webdir,
            far_trigger_cache=far_trigger_cache,
            seg_far_trigger_cache=seg_far_trigger_cache,
            post_pdf_cache=post_pdf_cache,
            marg_lr_prior_cache=marg_lr_prior_cache,
            mass_model_file=config.prior.mass_model,
            injections=bool(config.injections),
        )
        for layer in layer_list:
            dag.attach(layer)

        # triggers = dag.add_sim_inspiral_table(triggers)
        # merged_triggers, merged_inj_triggers =
        #           dag.find_injections_lite(merged_triggers)
        # dag.plot_summary(merged_triggers, pdfs)
        # dag.plot_background(merged_triggers, pdfs)
        # dag.plot_bin_background(dist_stats)
        # dag.plot_sensitivity(merged_triggers)
    else:
        raise ValueError(f"Unrecognized workflow: {args.workflow}")

    # Write dag and script to disk
    dag.write(pathlib.Path(args.dag_dir), write_script=True)
    dag.create_log_dir()


if __name__ == "__main__":
    main()
